/*
 * Armadillo-840 board support
 *
 * Copyright (C) 2012 Atmark Techno, Inc.
 * Copyright (C) 2012 Masahiro Nakai <nakai@atmark-techno.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include <linux/clk.h>
#include <linux/delay.h>
#include <linux/dma-mapping.h>
#include <linux/err.h>
#include <linux/gpio.h>
#include <linux/gpio_keys.h>
#include <linux/i2c-gpio.h>
#include <linux/input.h>
#include <linux/interrupt.h>
#include <linux/irq.h>
#include <linux/mmc/host.h>
#include <linux/mmc/sh_mobile_sdhi.h>
#include <linux/mfd/tmio.h>
#include <linux/platform_data/rmob-tpu-pwm.h>
#include <linux/platform_device.h>
#include <linux/pwm_backlight.h>
#include <linux/sh_eth.h>
#include <linux/sh_intc.h>
#include <linux/sh_timer.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/partitions.h>
#include <linux/mtd/physmap.h>
#include <linux/regulator/fixed.h>
#include <linux/regulator/machine.h>
#include <linux/rtc/rtc-s35390a.h>
#include <linux/slab.h>
#include <linux/spi/spi.h>
#include <linux/spi/sh_msiof.h>
#include <linux/usb.h>
#include <linux/usb/hcd.h>
#include <linux/usb/rmobile.h>
#include <linux/usb/renesas_usbhs.h>
#include <linux/videodev2.h>
#include <asm/mach-types.h>
#include <asm/mach/arch.h>
#include <asm/mach/time.h>
#include <asm/hardware/cache-l2x0.h>
#include <asm/hardware/gic.h>
#include <mach/common.h>
#include <mach/irqs.h>
#include <mach/r8a7740.h>
#include <media/ov772x.h>
#include <media/sh_mobile_ceu.h>
#include <media/soc_camera.h>
#include <sound/sh_fsi.h>
#include <sound/simple_card.h>
#include <video/sh_mobile_lcdc.h>
#include <video/sh_mobile_hdmi.h>
#include <video/sh_mobile_sdenc.h>

/*
 * Armadillo-840
 * =============
 *
 * CON1		SD
 * CON2		LAN
 * CON3		HDMI Output
 * CON4		UART
 * CON5		USB A Double Stack (Upper: USB0, Lower: USB1)
 * CON6		JTAG
 * CON7		Extension 1 (100pin)
 * CON8		Extension 2 (60pin)
 * CON9		Power Output
 * CON10	Power Input (DC Jack EIAJ #2)
 * CON11	Power Input
 * CON12	RTC BackUp for extend battery
 * LED1-LED2	User LED(Yellow, SMD)
 * SW1		Reset
 * JP1		HERMIT_EN
 *			open	OS auto boot
 *			short	Bootloader maintenance mode
 * JP2		JTAG_EN
 *			open	ARM JTAG diabled
 *			short	ARM JTAG enabled
 * JP3		SDBOOT_EN
 *			open	NOR boot
 *			short	SD boot (require for JP2 is open)
 *
 *
 * Armadillo-840 extension board
 * =============================
 *
 * CON1		Extension (100pin, connected to CON7 of Armadillo-840)
 * CON4		USB A
 * CON5		LCD (connected to AM-800480L1TMQW-T00H)
 * CON6		Extension LCD
 * CON7		Touch-Screen (connected to AM-800480L1TMQW-T00H)
 * CON8		Mic input
 * CON9		Headphone output
 * CON10	Audio-Line(L/R) / Composite-Video Output
 * CON11	Extension Audio
 * CON12	SD
 * CON13	AWL13 Module
 * CON14	Extension SD
 * CON15	Extension CAMERA
 * CON16	CAMERA (connected to KBCR-iC01VG)
 * CON17	Extension Switch
 * SW1-SW4	User programable (default Key via "gpio-key")
 * SW5		Reset
 * JP1-JP2	SDHI1 selector
 *		+--------+--------+------------------------------------+
 *		| JP2    | JP1    | SDHI1                              |
 *		+--------+--------+------------------------------------+
 *		| open   | open   | connected to CON12 SD              |
 *		| open   | short  | connected to CON13 AWL13           |
 *		| short  | open   | connected to CON14 Extension SD    |
 *		| short  | short  | Don't care                         |
 *		+--------+--------+------------------------------------+
 */

extern int gpio_reserve(unsigned offset);
extern int is_gpio_reserved(unsigned offset);

/* the 'con7eb_detected' indicate the extension board is connected to
 * CON7 of Armadillo-840. */
static int con7eb_detected;

static inline u32 portcr(unsigned nr)
{
	u32 offset = 0xE6050000;
	if (nr < 84)
		return offset + nr;
	if (nr < 115)
		return offset + 0x1000 + nr;
	if (nr < 210)
		return offset + 0x2000 + nr;
	return offset + 0x3000 + nr;
}

static int __maybe_unused gpio_request_array_simple(const struct gpio *array,
						    size_t num)
{
	int i, err;

	for (i = 0; i < num; i++, array++) {
		err = gpio_request(array->gpio, array->label);
		if (err)
			goto err_free;
	}
	return 0;

err_free:
	while (i--)
		gpio_free((--array)->gpio);
	return err;
}

/* ACM */
static struct platform_device __maybe_unused acm_device = {
	.name 	= "acm",
	.dev = {
		.coherent_dma_mask      = 0xffffffff,
	},
};

/* SGX540 */
static struct platform_device __maybe_unused sgx540_device = {
	.name		= "pvrsrvkm",
	.id		= -1,
};

/* UART.SCIFA/B */
static struct gpio __maybe_unused scifa0_gpios[] = {
	{ GPIO_FN_SCIFA0_RXD,		0,	"SCIFA0_RXD" }, /* PORT197 */
	{ GPIO_FN_SCIFA0_TXD,		0,	"SCIFA0_TXD" }, /* PORT198 */
};

static struct gpio __maybe_unused scifa0_rtscts_gpios[] = {
	{ GPIO_FN_SCIFA0_RTS,		0,	"SCIFA0_RTS" }, /* PORT194 */
	{ GPIO_FN_SCIFA0_CTS,		0,	"SCIFA0_CTS" }, /* PORT193 */
};

static struct gpio __maybe_unused scifa1_gpios[] = {
	{ GPIO_FN_SCIFA1_RXD,		0,	"SCIFA1_RXD" }, /* PORT195 */
	{ GPIO_FN_SCIFA1_TXD,		0,	"SCIFA1_TXD" }, /* PORT196 */
};

static struct gpio __maybe_unused scifa1_rtscts_gpios[] = {
	{ GPIO_FN_SCIFA1_RTS,		0,	"SCIFA1_RTS" }, /* PORT23 */
	{ GPIO_FN_SCIFA1_CTS,		0,	"SCIFA1_CTS" }, /* PORT21 */
};

static struct gpio scifa2_gpios[] = {
	{ GPIO_FN_SCIFA2_RXD,		0,	"SCIFA2_RXD" }, /* PORT201 */
	{ GPIO_FN_SCIFA2_TXD,		0,	"SCIFA2_TXD" }, /* PORT200 */
	{ GPIO_FN_SCIFA2_RTS,		0,	"SCIFA2_RTS" }, /* PORT96 */
	{ GPIO_FN_SCIFA2_CTS,		0,	"SCIFA2_CTS" }, /* PORT95 */
};

static struct gpio __maybe_unused scifa3_gpios[] = {
	{ GPIO_FN_SCIFA3_RXD_PORT159,	0,	"SCIFA3_RXD" },
	{ GPIO_FN_SCIFA3_TXD_PORT160,	0,	"SCIFA3_TXD" },
};

static struct gpio __maybe_unused scifa4_12_13_gpios[] = {
	{ GPIO_FN_SCIFA4_RXD_PORT12,	0,	"SCIFA4_RXD" },
	{ GPIO_FN_SCIFA4_TXD_PORT13,	0,	"SCIFA4_TXD" },
};

static struct gpio __maybe_unused scifa4_94_93_gpios[] = {
	{ GPIO_FN_SCIFA4_RXD_PORT94,	0,	"SCIFA4_RXD" },
	{ GPIO_FN_SCIFA4_TXD_PORT93,	0,	"SCIFA4_TXD" },
};

static struct gpio __maybe_unused scifa5_gpios[] = {
	{ GPIO_FN_SCIFA5_RXD_PORT10,	0,	"SCIFA5_RXD" },
	{ GPIO_FN_SCIFA5_TXD_PORT20,	0,	"SCIFA5_TXD" },
};

static struct gpio __maybe_unused scifa6_gpios[] = {
	{ GPIO_FN_SCIFA6_RXD,		0,	"SCIFA6_RXD" }, /* PORT25 */
	{ GPIO_FN_SCIFA6_TXD,		0,	"SCIFA6_TXD" }, /* PORT26 */
};

static struct gpio __maybe_unused scifa7_gpios[] = {
	{ GPIO_FN_SCIFA7_RXD,		0,	"SCIFA7_RXD" }, /* PORT0 */
	{ GPIO_FN_SCIFA7_TXD,		0,	"SCIFA7_TXD" }, /* PORT1 */
};

static struct gpio __maybe_unused scifb_gpios[] = {
	{ GPIO_FN_SCIFB_RXD_PORT3,	0,	"SCIFB_RXD" },
	{ GPIO_FN_SCIFB_TXD_PORT4,	0,	"SCIFB_TXD" },
};

static struct gpio __maybe_unused scifb_rtscts_gpios[] = {
	{ GPIO_FN_SCIFB_RTS_PORT172,	0,	"SCIFB_RTS" },
	{ GPIO_FN_SCIFB_CTS_PORT173,	0,	"SCIFB_CTS" },
};

static void a840_scifx_init(void)
{
	int ret;

	/* SCIFA2 (CON3) */
	gpio_request_array_simple(scifa2_gpios, ARRAY_SIZE(scifa2_gpios));

	if (IS_ENABLED(CONFIG_ARMADILLO840_SCIFA0)) {
		ret = gpio_request_array_simple(scifa0_gpios,
						ARRAY_SIZE(scifa0_gpios));
		if (!ret) {
			gpio_request_pullup(portcr(197)); /* RXD */
			gpio_request_pulloff(portcr(198)); /* TXD */
		}
		if (!ret && IS_ENABLED(CONFIG_ARMADILLO840_SCIFA0_HAVE_RTSCTS)) {
			ret = gpio_request_array_simple(scifa0_rtscts_gpios,
							ARRAY_SIZE(scifa0_rtscts_gpios));
			if (!ret) {
				gpio_request_pulldown(portcr(194)); /* RTS */
				gpio_request_pullup(portcr(193)); /* CTS */
			}
		}
		if (ret)
			pr_err("Any SCIFA0 resources are already reserved.\n");
	}

	if (IS_ENABLED(CONFIG_ARMADILLO840_SCIFA1)) {
		ret = gpio_request_array_simple(scifa1_gpios,
						ARRAY_SIZE(scifa1_gpios));
		if (!ret) {
			gpio_request_pullup(portcr(195)); /* RXD */
			gpio_request_pulloff(portcr(196)); /* TXD */
		}
		if (!ret && IS_ENABLED(CONFIG_ARMADILLO840_SCIFA1_HAVE_RTSCTS)) {
			ret = gpio_request_array_simple(scifa1_rtscts_gpios,
							ARRAY_SIZE(scifa1_rtscts_gpios));
			if (!ret) {
				gpio_request_pulldown(portcr(23)); /* RTS */
				gpio_request_pullup(portcr(21)); /* CTS */
			}
		}
		if (ret)
			pr_err("Any SCIFA1 resources are already reserved.\n");
	}

	if (IS_ENABLED(CONFIG_ARMADILLO840_SCIFA3)) {
		ret = gpio_request_array_simple(scifa3_gpios,
						ARRAY_SIZE(scifa3_gpios));
		if (!ret) {
			gpio_request_pullup(portcr(159)); /* RXD */
			gpio_request_pulloff(portcr(160)); /* TXD */
		} else
			pr_err("Any SCIFA3 resources are already reserved.\n");
	}

	if (IS_ENABLED(CONFIG_ARMADILLO840_SCIFA4_12_13)) {
		ret = gpio_request_array_simple(scifa4_12_13_gpios,
						ARRAY_SIZE(scifa4_12_13_gpios));
		if (!ret) {
			gpio_request_pullup(portcr(12)); /* RXD */
			gpio_request_pulloff(portcr(13)); /* TXD */
		} else
			pr_err("Any SCIFA4 resources are already reserved.\n");
	}

	if (IS_ENABLED(CONFIG_ARMADILLO840_SCIFA4_94_93)) {
		ret = gpio_request_array_simple(scifa4_94_93_gpios,
						ARRAY_SIZE(scifa4_94_93_gpios));
		if (!ret) {
			gpio_request_pullup(portcr(94)); /* RXD */
			gpio_request_pulloff(portcr(93)); /* TXD */
		} else
			pr_err("Any SCIFA4 resources are already reserved.\n");
	}

	if (IS_ENABLED(CONFIG_ARMADILLO840_SCIFA5)) {
		ret = gpio_request_array_simple(scifa5_gpios,
						ARRAY_SIZE(scifa5_gpios));
		if (!ret) {
			gpio_request_pullup(portcr(10)); /* RXD */
			gpio_request_pulloff(portcr(20)); /* TXD */
		} else
			pr_err("Any SCIFA5 resources are already reserved.\n");
	}

	if (IS_ENABLED(CONFIG_ARMADILLO840_SCIFA6)) {
		ret = gpio_request_array_simple(scifa6_gpios,
						ARRAY_SIZE(scifa6_gpios));
		if (!ret) {
			gpio_request_pullup(portcr(25)); /* RXD */
			gpio_request_pulloff(portcr(26)); /* TXD */
		} else
			pr_err("Any SCIFA6 resources are already reserved.\n");
	}

	if (IS_ENABLED(CONFIG_ARMADILLO840_SCIFA7)) {
		ret = gpio_request_array_simple(scifa7_gpios,
						ARRAY_SIZE(scifa7_gpios));
		if (!ret) {
			gpio_request_pullup(portcr(0)); /* RXD */
			gpio_request_pulloff(portcr(1)); /* TXD */
		} else
			pr_err("Any SCIFA7 resources are already reserved.\n");
	}

	if (IS_ENABLED(CONFIG_ARMADILLO840_SCIFB)) {
		ret = gpio_request_array_simple(scifb_gpios,
						ARRAY_SIZE(scifb_gpios));
		if (!ret) {
			gpio_request_pullup(portcr(3)); /* RXD */
			gpio_request_pulloff(portcr(4)); /* TXD */
		}
		if (!ret && IS_ENABLED(CONFIG_ARMADILLO840_SCIFB_HAVE_RTSCTS)) {
			ret = gpio_request_array_simple(scifb_rtscts_gpios,
							ARRAY_SIZE(scifb_rtscts_gpios));
			if (!ret) {
				gpio_request_pulldown(portcr(172)); /* RTS */
				gpio_request_pullup(portcr(173)); /* CTS */
			}
		}
		if (ret)
			pr_err("Any SCIFB resources are already reserved.\n");
	}
}

/* SPI.MSIOF */
static struct spi_board_info __maybe_unused msiof1_board_info[] __initdata = {
	/* this is inplementation dependent. */
	{
		.modalias		= "spidev",
		.controller_data	= (void *)GPIO_PORT73,
		.max_speed_hz		= 1000000,
		.bus_num		= 1,
		.chip_select		= 0,
	}
};

static struct sh_msiof_spi_info __maybe_unused sh_msiof1_info = {
	.rx_fifo_override	= 256,
	.num_chipselect		= 3,
};

static struct resource __maybe_unused sh_msiof1_resources[] = {
	[0] = {
		.start		= 0xE6E10000,
		.end		= 0xE6E10064 - 1,
		.flags		= IORESOURCE_MEM,
	},
	[1] = {
		.start		= gic_spi(91),
		.flags		= IORESOURCE_IRQ,
	},
};

static struct platform_device __maybe_unused sh_msiof1_device = {
	.name		= "spi_sh_msiof",
	.id		= 1,
	.dev		= {
		.platform_data	= &sh_msiof1_info,
	},
	.resource	= sh_msiof1_resources,
	.num_resources	= ARRAY_SIZE(sh_msiof1_resources),
};

static struct gpio __maybe_unused msiof1_gpios[] = {
	{ GPIO_FN_MSIOF1_TSCK_PORT72,	0,	"TSCK" },
	{ GPIO_FN_MSIOF1_TXD_PORT74,	0,	"MISO" },
	{ GPIO_FN_MSIOF1_RXD_PORT75,	0,	"MOSI" },
};

static void a840_msiof_init(void)
{
	if (!IS_ENABLED(CONFIG_ARMADILLO840_MSIOF1))
		return;

	if (gpio_request_array_simple(msiof1_gpios, ARRAY_SIZE(msiof1_gpios))) {
		pr_err("Any MSIOF1 resources are already reserved.\n");
		return;
	}

	platform_device_register(&sh_msiof1_device);

	if (IS_ENABLED(CONFIG_ARMADILLO840_MSIOF1_DEVICE_SPIDEV_PORT73)) {
		gpio_request(GPIO_PORT73, NULL); /* CS */
		gpio_direction_output(GPIO_PORT73, 0);

		spi_register_board_info(msiof1_board_info,
					ARRAY_SIZE(msiof1_board_info));
	}
}

/* LAN */
static struct sh_eth_plat_data sh_eth_platdata = {
	.phy			= 0x00, /* LAN8710A */
	.edmac_endian		= EDMAC_LITTLE_ENDIAN,
	.register_type		= SH_ETH_REG_GIGABIT,
	.phy_interface		= PHY_INTERFACE_MODE_MII,
};

static struct resource sh_eth_resources[] = {
	{
		.start	= 0xe9a00000,
		.end	= 0xe9a00800 - 1,
		.flags	= IORESOURCE_MEM,
	}, {
		.start	= 0xe9a01800,
		.end	= 0xe9a02000 - 1,
		.flags	= IORESOURCE_MEM,
	}, {
		.start	= gic_spi(110),
		.flags	= IORESOURCE_IRQ,
	},
};

static struct platform_device sh_eth_device = {
	.name = "sh-eth",
	.id = -1,
	.dev = {
		.platform_data = &sh_eth_platdata,
	},
	.resource = sh_eth_resources,
	.num_resources = ARRAY_SIZE(sh_eth_resources),
};

static struct gpio eth_gpios[] = {
	{ GPIO_FN_ET_CRS,		0,	"ET_CRS" },
	{ GPIO_FN_ET_MDC,		0,	"ET_MDC" },
	{ GPIO_FN_ET_MDIO,		0,	"ET_MDIO" },
	{ GPIO_FN_ET_TX_ER,		0,	"ET_TX_ER" },
	{ GPIO_FN_ET_RX_ER,		0,	"ET_RX_ER" },
	{ GPIO_FN_ET_ERXD0,		0,	"ET_ERXD0" },
	{ GPIO_FN_ET_ERXD1,		0,	"ET_ERXD1" },
	{ GPIO_FN_ET_ERXD2,		0,	"ET_ERXD2" },
	{ GPIO_FN_ET_ERXD3,		0,	"ET_ERXD3" },
	{ GPIO_FN_ET_TX_CLK,		0,	"ET_TX_CLK" },
	{ GPIO_FN_ET_TX_EN,		0,	"ET_TX_EN" },
	{ GPIO_FN_ET_ETXD0,		0,	"ET_ETXD0" },
	{ GPIO_FN_ET_ETXD1,		0,	"ET_ETXD1" },
	{ GPIO_FN_ET_ETXD2,		0,	"ET_ETXD2" },
	{ GPIO_FN_ET_ETXD3,		0,	"ET_ETXD3" },
	{ GPIO_FN_ET_COL,		0,	"ET_COL" },
	{ GPIO_FN_ET_RX_DV,		0,	"ET_RX_DV" },
	{ GPIO_FN_ET_RX_CLK,		0,	"ET_RX_CLK" },
	{ GPIO_PORT208,			0,	"PHY_RST" },
};

static void a840_eth_init(void)
{
	gpio_request_array_simple(eth_gpios, ARRAY_SIZE(eth_gpios));

	/* reset phy */
	gpio_direction_output(GPIO_PORT208, 0);
	mdelay(10);
	gpio_direction_output(GPIO_PORT208, 1);
}

/* SDHI */
static struct regulator_consumer_supply sdhi0_supply[] = {
	REGULATOR_SUPPLY("vmmc", "sh_mobile_sdhi.0"),
};

static struct regulator_init_data sdhi0_regulator_init_data = {
	.constraints = {
		.max_uA		= 500000,
		.valid_ops_mask	= REGULATOR_CHANGE_STATUS,
	},
	.num_consumer_supplies	= ARRAY_SIZE(sdhi0_supply),
	.consumer_supplies	= sdhi0_supply,
};

static struct fixed_voltage_config sdhi0_regulator_config = {
	.supply_name		= "sdhi0",
	.gpio			= GPIO_PORT166,
	.enable_high		= 1,
	.enabled_at_boot	= 0,
	.startup_delay		= 100000,
	.microvolts		= 3300000,
	.init_data		= &sdhi0_regulator_init_data,
};

static struct platform_device sdhi0_regulator = {
	.name		= "reg-fixed-voltage",
	.id		= 0,
	.dev		= {
		.platform_data	= &sdhi0_regulator_config,
	},
};

static struct sh_mobile_sdhi_info sdhi0_info = {
	.dma_slave_tx	= SHDMA_SLAVE_SDHI0_TX,
	.dma_slave_rx	= SHDMA_SLAVE_SDHI0_RX,
	.tmio_caps	= MMC_CAP_SD_HIGHSPEED | MMC_CAP_SDIO_IRQ,
	.tmio_ocr_mask	= MMC_VDD_165_195 | MMC_VDD_32_33 | MMC_VDD_33_34,
};

static struct resource sdhi0_resources[] = {
	{
		.name	= "SDHI0",
		.start	= 0xe6850000,
		.end	= 0xe6850100 - 1,
		.flags	= IORESOURCE_MEM,
	},
	{
		.name	= SH_MOBILE_SDHI_IRQ_CARD_DETECT,
		.start	= gic_spi(117),
		.flags	= IORESOURCE_IRQ,
	},
	{
		.name	= SH_MOBILE_SDHI_IRQ_SDCARD,
		.start	= gic_spi(118),
		.flags	= IORESOURCE_IRQ,
	},
	{
		.name	= SH_MOBILE_SDHI_IRQ_SDIO,
		.start	= gic_spi(119),
		.flags	= IORESOURCE_IRQ,
	},
};

static struct platform_device sdhi0_device = {
	.name		= "sh_mobile_sdhi",
	.id		= 0,
	.dev		= {
		.platform_data	= &sdhi0_info,
	},
	.num_resources	= ARRAY_SIZE(sdhi0_resources),
	.resource	= sdhi0_resources,
};

static struct gpio __maybe_unused sdhi0_gpios[] = {
	{ GPIO_FN_SDHI0_CMD,	0,	"SDHI0_CMD" },
	{ GPIO_FN_SDHI0_CLK,	0,	"SDHI0_CLK" },
	{ GPIO_FN_SDHI0_D0,	0,	"SDHI0_D0" },
	{ GPIO_FN_SDHI0_D1,	0,	"SDHI0_D1" },
	{ GPIO_FN_SDHI0_D2,	0,	"SDHI0_D2" },
	{ GPIO_FN_SDHI0_D3,	0,	"SDHI0_D3" },
	{ GPIO_FN_SDHI0_WP,	0,	"SDHI0_WP" },
	{ GPIO_FN_SDHI0_CD,	0,	"SDHI0_CD" },
};

static struct regulator_consumer_supply __maybe_unused sdhi1_supply[] = {
	REGULATOR_SUPPLY("vmmc", "sh_mobile_sdhi.1"),
};

static struct regulator_init_data __maybe_unused sdhi1_regulator_init_data = {
	.constraints = {
		.max_uA		= 500000,
		.valid_ops_mask	= REGULATOR_CHANGE_STATUS,
	},
	.num_consumer_supplies	= ARRAY_SIZE(sdhi1_supply),
	.consumer_supplies	= sdhi1_supply,
};

static struct fixed_voltage_config __maybe_unused sdhi1_regulator_config = {
	.supply_name		= "sdhi1",
	.gpio			= GPIO_PORT74,
	.enable_high		= 1,
	.enabled_at_boot	= 0,
	.startup_delay		= 100000,
	.microvolts		= 3300000,
	.init_data		= &sdhi1_regulator_init_data,
};

static struct platform_device __maybe_unused sdhi1_regulator = {
	.name		= "reg-fixed-voltage",
	.id		= 1,
	.dev		= {
		.platform_data	= &sdhi1_regulator_config,
	},
};

static struct sh_mobile_sdhi_info __maybe_unused sdhi1_info = {
	.dma_slave_tx	= SHDMA_SLAVE_SDHI1_TX,
	.dma_slave_rx	= SHDMA_SLAVE_SDHI1_RX,
	.tmio_caps	= MMC_CAP_SD_HIGHSPEED | MMC_CAP_SDIO_IRQ,
	.tmio_ocr_mask	= MMC_VDD_165_195 | MMC_VDD_32_33 | MMC_VDD_33_34,
};

static struct resource __maybe_unused sdhi1_resources[] = {
	{
		.name	= "SDHI1",
		.start	= 0xe6860000,
		.end	= 0xe6860100 - 1,
		.flags	= IORESOURCE_MEM,
	},
	{
		.name	= SH_MOBILE_SDHI_IRQ_CARD_DETECT,
		.start	= gic_spi(121),
		.flags	= IORESOURCE_IRQ,
	},
	{
		.name	= SH_MOBILE_SDHI_IRQ_SDCARD,
		.start	= gic_spi(122),
		.flags	= IORESOURCE_IRQ,
	},
	{
		.name	= SH_MOBILE_SDHI_IRQ_SDIO,
		.start	= gic_spi(123),
		.flags	= IORESOURCE_IRQ,
	},
};

static struct platform_device __maybe_unused sdhi1_device = {
	.name		= "sh_mobile_sdhi",
	.id		= 1,
	.dev		= {
		.platform_data	= &sdhi1_info,
	},
	.num_resources	= ARRAY_SIZE(sdhi1_resources),
	.resource	= sdhi1_resources,
};

static struct gpio __maybe_unused sdhi1_gpios[] = {
	{ GPIO_FN_SDHI1_CMD,	0,	"SDHI1_CMD" },
	{ GPIO_FN_SDHI1_CLK,	0,	"SDHI1_CLK" },
	{ GPIO_FN_SDHI1_D0,	0,	"SDHI1_D0" },
	{ GPIO_FN_SDHI1_D1,	0,	"SDHI1_D1" },
	{ GPIO_FN_SDHI1_D2,	0,	"SDHI1_D2" },
	{ GPIO_FN_SDHI1_D3,	0,	"SDHI1_D3" },
};

static struct gpio __maybe_unused sdhi1_cdwp_gpios[] = {
	{ GPIO_FN_SDHI1_WP,	0,	"SDHI1_WP" },
	{ GPIO_FN_SDHI1_CD,	0,	"SDHI1_CD" },
};

static void a840_sdhi_init(void)
{
	int ret;
	int i;

	/* SDHI0 */
	gpio_request_array_simple(sdhi0_gpios, ARRAY_SIZE(sdhi0_gpios));
	/* PORT76-83 terminations are handled on a extention board.
	 * remove the internal pull-up/down. */
	for (i = 76; i <= 83; i++)
		gpio_request_pulloff(portcr(i));

	/* SDHI0 power is controlled by PORT166 on extention board */
	gpio_reserve(GPIO_PORT166);
	platform_device_register(&sdhi0_regulator);

	platform_device_register(&sdhi0_device);

	if (con7eb_detected && IS_ENABLED(CONFIG_ARMADILLO840_SDHI1)) {
		ret = gpio_request_array_simple(sdhi1_gpios,
						ARRAY_SIZE(sdhi1_gpios));
		if (!ret && IS_ENABLED(CONFIG_ARMADILLO840_SDHI1_HAVE_CDWP_72_73))
			ret = gpio_request_array_simple(sdhi1_cdwp_gpios,
							ARRAY_SIZE(sdhi1_cdwp_gpios));
		if (ret) {
			pr_err("Any SDHI1 resources are already reserved.\n");
			return;
		}
		/* PORT66-73 terminations are handled on a extention board.
		 * remove the internal pull-up/down. */
		for (i = 66; i <= 71; i++)
			gpio_request_pulloff(portcr(i));
		if (IS_ENABLED(CONFIG_ARMADILLO840_SDHI1_HAVE_CDWP_72_73)) {
			gpio_request_pulloff(portcr(72));
			gpio_request_pulloff(portcr(73));
		}

		/* if your board doesn't use CD pin of SDHI1, you can enable
		 * polling to detect card by itself. */
		if (!IS_ENABLED(CONFIG_ARMADILLO840_SDHI1_HAVE_CDWP_72_73))
			sdhi1_info.tmio_caps |= MMC_CAP_NEEDS_POLL;

		if (IS_ENABLED(CONFIG_ARMADILLO840_SDHI1_HAVE_PWRSW_PORT74)) {
			/* SDHI1 power is controlled by PORT74 on
			 * extention board */
			gpio_reserve(GPIO_PORT74);
			platform_device_register(&sdhi1_regulator);
		}

		platform_device_register(&sdhi1_device);
	}
}

/* USB Port1 mode */
static int usb_port1_is_device = 0;

/* USB Port1 connector */
static int usb_port1_is_connected_to_con5 = 0;

/* USB */
#define USBCKCR		0xe615008c
#define USBCR1		0xe605810a

#define USBH		0xC6700000
#define USBH_USBCTR	0x10834

static DEFINE_SPINLOCK(usb_lock);

struct r8a7740_usb_clock_priv {
	const char *name;
	struct list_head list;
	int use_count;

	struct clk *phy;
	struct clk *usb24;
	struct clk *pci;
	struct clk *func;
	struct clk *host;
};
#define to_r8a7740_usb_clock_priv(list) \
	container_of(list, struct r8a7740_usb_clock_priv, list)

static LIST_HEAD(r8a7740_usb_clock_list);
static void usb_common_clock(struct platform_device *pdev, int on)
{
	static void *host_base;
	struct r8a7740_usb_clock_priv *tmp, *priv = NULL;
	struct list_head *list;
	int use_count = 0;
	unsigned long flags;

	spin_lock_irqsave(&usb_lock, flags);

	list_for_each(list, &r8a7740_usb_clock_list) {
		tmp = to_r8a7740_usb_clock_priv(list);
		if (strcmp(dev_name(&pdev->dev), tmp->name) == 0)
			priv = tmp;
		use_count += tmp->use_count;
	}
	if (!on && use_count == 0) {
		spin_unlock_irqrestore(&usb_lock, flags);
		return;
	}

	if (!priv) {
		/* create clock data */
		priv = kzalloc(sizeof(struct r8a7740_usb_clock_priv),
			       GFP_KERNEL);
		priv->name = dev_name(&pdev->dev);
		list_add_tail(&priv->list, &r8a7740_usb_clock_list);
	}

	if (on && priv->use_count++ == 0) {
		priv->phy	= clk_get(&pdev->dev, "phy");
		priv->usb24	= clk_get(&pdev->dev, "usb24");
		priv->pci	= clk_get(&pdev->dev, "pci");
		priv->func	= clk_get(&pdev->dev, "func");
		priv->host	= clk_get(&pdev->dev, "host");
		clk_enable(priv->usb24);
		clk_enable(priv->pci);
		clk_enable(priv->host);
		clk_enable(priv->func);
		clk_enable(priv->phy);
	}

	if (on && use_count == 0) {
		host_base = ioremap_nocache(USBH, 0x20000);
		if (!usb_port1_is_device) {
			/* Port1:HOST Port2:HOST */
			__raw_writew(0x6b50, USBCR1);
		} else {
			/* Port1:Function Port2:HOST
			 * USBCR.P1PORTSEL = 11b(Port1 is driven by UFB FUNC)
			 * USBCR.P2PORTSEL = 01b(Port2 is driven by USB HOST)
			 * USBCR.PORTNUM   = 01b(One HOST)
			 */
			__raw_writew(0xd750, USBCR1);
		}
		mdelay(1);

		__raw_writel(0x0000040c, host_base + USBH_USBCTR);
		__raw_writel(0x00000408, host_base + USBH_USBCTR);
		mdelay(10);

		if (usb_port1_is_device)
			__raw_writew(0xd770, USBCR1);
	}

	if (!on && use_count - 1 == 0) {
		__raw_writel(0x0000050e, host_base + USBH_USBCTR);
		__raw_writew(0x07c0, USBCR1);
		iounmap(host_base);
	}

	if (!on && --priv->use_count == 0) {
		clk_disable(priv->phy);
		clk_disable(priv->func);	/* usb work around */
		clk_disable(priv->host);	/* usb work around */
		clk_disable(priv->pci);		/* usb work around */
		clk_disable(priv->usb24);	/* usb work around */
		clk_put(priv->phy);
		clk_put(priv->usb24);
		clk_put(priv->pci);
		clk_put(priv->host);
		clk_put(priv->func);
		list_del(&priv->list);
		kfree(priv);
	}

	spin_unlock_irqrestore(&usb_lock, flags);
}

static void usb_host_set_vbus(struct platform_device *pdev, int on)
{
	struct usb_hcd *hcd = platform_get_drvdata(pdev);

	switch (hcd->self.busnum) {
	case 1:
		gpio_set_value(GPIO_PORT85, !!on);
		break;
	case 2:
		gpio_set_value(GPIO_PORT87, !!on);
		break;
	default:
		break;
	}
	msleep(10);
}

static struct rmobile_usb_platdata usb_host_platdata = {
	.set_vbus	= usb_host_set_vbus,
	.common_clock	= usb_common_clock,
};

static u64 usb_host_dma_mask = DMA_BIT_MASK(32);

static struct resource usb_ohci_resources[] = {
	[0] = {
		.start	= 0xc6700000,
		.end	= 0xc6720000 - 1,
		.flags	= IORESOURCE_MEM,
	},
	[1] = {
		.start	= gic_spi(234),
		.flags	= IORESOURCE_IRQ,
	},
};

static struct platform_device usb_ohci_device = {
	.name		= "rmobile-ohci-driver",
	.id		= -1,
	.dev		= {
		.dma_mask		= &usb_host_dma_mask,
		.coherent_dma_mask	= DMA_BIT_MASK(32),
		.platform_data		= &usb_host_platdata,
	},
	.num_resources	= ARRAY_SIZE(usb_ohci_resources),
	.resource	= usb_ohci_resources,
};

static struct resource usb_ehci_resources[] = {
	[0] = {
		.start	= 0xc6700000,
		.end	= 0xc6720000 - 1,
		.flags	= IORESOURCE_MEM,
	},
	[1] = {
		.start	= gic_spi(234),
		.flags	= IORESOURCE_IRQ,
	},
};

static struct platform_device usb_ehci_device = {
	.name		= "rmobile-ehci-driver",
	.id		= -1,
	.dev		= {
		.dma_mask		= &usb_host_dma_mask,
		.coherent_dma_mask	= DMA_BIT_MASK(32),
		.platform_data		= &usb_host_platdata,
	},
	.num_resources	= ARRAY_SIZE(usb_ehci_resources),
	.resource	= usb_ehci_resources,
};

/*
 * USB function
 *
 * USBF needs workaround on R8A7740 chip.
 * These are a little bit complex.
 * see
 *	usbhsf_power_ctrl()
 */
/*
 * VBUS setting
 *
 * this is implementation dependent.
 * for example if PORT209 will be assigned to detect VBUS
 * static unsigned usb_port1_vbus_gpio		= GPIO_PORT209;
 * static unsigned usb_port1_vbus_irqgpio	= GPIO_FN_IRQ7_PORT209;
 * static int usb_port1_vbus_irq		= irq_pin(7);
 */
static unsigned usb_port1_vbus_gpio	= GPIO_PORT20;
static unsigned usb_port1_vbus_irqgpio	= GPIO_FN_IRQ1;
static int usb_port1_vbus_irq		= irq_pin(1);

struct usbhsf_private {
	struct clk *phy;
	struct clk *usb24;
	struct clk *pci;
	struct clk *func;
	struct clk *host;
	void __iomem *usbh_base;
	struct renesas_usbhs_platform_info info;
};

#define usbhsf_get_priv(pdev)				\
	container_of(renesas_usbhs_get_info(pdev),	\
		     struct usbhsf_private, info)

static int usbhsf_get_id(struct platform_device *pdev)
{
	return USBHS_GADGET;
}

static void usbhsf_power_ctrl(struct platform_device *pdev,
			      void __iomem *base, int enable)
{
	/*
	 * Work around for USB Function.
	 * It needs USB host clock, and settings
	 */
	usb_common_clock(pdev, enable);

	if (enable) {
		/*
		 * USB PHY Power ON
		 */
		__raw_writew(0xd770, USBCR1);
		__raw_writew(0x4000, base + 0x102); /* USBF :: SUSPMODE */
	}
}

static int usbhsf_get_vbus(struct platform_device *pdev)
{
	return gpio_get_value(usb_port1_vbus_gpio);
}

static irqreturn_t usbhsf_interrupt(int irq, void *data)
{
	struct platform_device *pdev = data;

	renesas_usbhs_call_notify_hotplug(pdev);

	return IRQ_HANDLED;
}

static void usbhsf_hardware_exit(struct platform_device *pdev)
{
	free_irq(usb_port1_vbus_irq, pdev);
}

static int usbhsf_hardware_init(struct platform_device *pdev)
{
	struct clk *usb24_clk;
	int ret;

	ret = request_irq(usb_port1_vbus_irq, usbhsf_interrupt,
			  IRQF_TRIGGER_NONE, dev_name(&pdev->dev), pdev);
	if (ret) {
		dev_err(&pdev->dev, "request_irq err\n");
		return ret;
	}
	irq_set_irq_type(usb_port1_vbus_irq, IRQ_TYPE_EDGE_BOTH);

	/* usb24 use 1/1 of parent clock (= usb24s = 24MHz) */
	usb24_clk = clk_get(&pdev->dev, "usb24");
	clk_set_rate(usb24_clk,
		     clk_get_rate(clk_get_parent(usb24_clk)));
	clk_put(usb24_clk);

	return 0;
}

static struct usbhsf_private usbhsf_private = {
	.info = {
		.platform_callback = {
			.get_id		= usbhsf_get_id,
			.get_vbus	= usbhsf_get_vbus,
			.hardware_init	= usbhsf_hardware_init,
			.hardware_exit	= usbhsf_hardware_exit,
			.power_ctrl	= usbhsf_power_ctrl,
		},
		.driver_param = {
			.buswait_bwait		= 5,
			.detection_delay	= 5,
			/* .d0_rx_id	= SHDMA_SLAVE_USBHS_RX, */
			/* .d1_tx_id	= SHDMA_SLAVE_USBHS_TX, */
		},
	}
};

static struct resource usbhsf_resources[] = {
	{
		.name	= "USBHS",
		.start	= 0xe6890000,
		.end	= 0xe6890104 - 1,
		.flags	= IORESOURCE_MEM,
	},
	{
		.start	= gic_spi(51),
		.flags	= IORESOURCE_IRQ,
	},
};

static struct platform_device usbhsf_device = {
	.name		= "renesas_usbhs",
	.dev		= {
		.platform_data	= &usbhsf_private.info,
	},
	.id		= -1,
	.num_resources	= ARRAY_SIZE(usbhsf_resources),
	.resource	= usbhsf_resources,
};

static struct gpio usb_gpios[] = {
	{ GPIO_PORT85,			0,	"USB0_PPON" },
	{ GPIO_FN_USB0_OCI,		0,	"USB0_OCI" },
	{ GPIO_PORT87,			0,	"USB1_PPON" },
	{ GPIO_FN_USB1_OCI,		0,	"USB1_OCI" },
	{ GPIO_PORT176,			0,	"EXT_USB_EN"},
};

static void a840_usb_init(void)
{
	gpio_request_array_simple(usb_gpios, ARRAY_SIZE(usb_gpios));
	gpio_direction_output(GPIO_PORT85, 0);
	gpio_direction_output(GPIO_PORT87, 0);

	writel(readl(USBCKCR) | 0x40, USBCKCR); /* USB div output */

	if (IS_ENABLED(CONFIG_ARMADILLO840_USB1_CON7_DEVICE))
		usb_port1_is_device = 1;
	if (IS_ENABLED(CONFIG_ARMADILLO840_USB1_CON5_HOST))
		usb_port1_is_connected_to_con5 = 1;

	/* Armadillo-840 can select where USB Port1
	 * is connected to via the USB bus switch on the board.
	 * CON5 if PORT176 is Low, CON7 otherwise.
	 */
	if (usb_port1_is_connected_to_con5)
		gpio_direction_output(GPIO_PORT176, 0);
	else
		gpio_direction_output(GPIO_PORT176, 1);

	if (usb_port1_is_device &&
	    usb_port1_vbus_gpio == 0 &&
	    usb_port1_vbus_irqgpio == 0 &&
	    usb_port1_vbus_irq == 0) {
		printk(KERN_INFO "Looks like no gpio is assigned to detect VBUS of USB Port1.\n"
		       "We are disabling USB Port1 for now.");
	} else if (usb_port1_is_device) {
		/* A1 chip has 2 IRQ7 pin and it was controled by MSEL register.
		 * OTOH, usbhs interrupt needs its value (HI/LOW) to decide
		 * USB connection/disconnection (usbhsf_get_vbus()).
		 * For exapmle if PORT209 is assigned to detect VBUS,
		 * this means we needs to select GPIO_FN_IRQ7_PORT209 first,
		 * and select GPIO_PORT209 here
		 */
		gpio_request(usb_port1_vbus_irqgpio, NULL);
		gpio_request(usb_port1_vbus_gpio, NULL);
		gpio_direction_input(usb_port1_vbus_gpio);
		irq_set_irq_type(usb_port1_vbus_irq, IRQ_TYPE_EDGE_BOTH);

		platform_device_register(&usbhsf_device);
	}
}

/* FSI */
static int fsi_hdmi_set_rate(struct device *dev, int rate, int enable)
{
	/* it support 48KHz only */
	if (48000 != rate)
		return -EINVAL;

	return SH_FSI_ACKMD_256 | SH_FSI_BPFMD_64;
}

static struct sh_fsi_platform_info fsi_info = {
	.port_a = {
		.flags		= SH_FSI_BRS_INV,
		.tx_id		= SHDMA_SLAVE_FSIA_TX,
	},
	/* FSI-HDMI */
	.port_b = {
		.flags		= SH_FSI_FMT_SPDIF |
				  SH_FSI_ENABLE_STREAM_MODE,
		.set_rate	= fsi_hdmi_set_rate,
		.tx_id		= SHDMA_SLAVE_FSIB_TX,
	}
};

static struct resource fsi_resources[] = {
	[0] = {
		.name	= "FSI",
		.start	= 0xfe1f0000,
		.end	= 0xfe1f8400 - 1,
		.flags	= IORESOURCE_MEM,
	},
	[1] = {
		.start  = gic_spi(9),
		.flags  = IORESOURCE_IRQ,
	},
};

static struct platform_device fsi_device = {
	.name		= "sh_fsi2",
	.id		= -1,
	.num_resources	= ARRAY_SIZE(fsi_resources),
	.resource	= fsi_resources,
	.dev	= {
		.platform_data	= &fsi_info,
	},
};

static struct gpio __maybe_unused fsia_gpios[] = {
	{ GPIO_FN_FSIAIBT,		0,	"FSIAIBT" },
	{ GPIO_FN_FSIAILR,		0,	"FSIAILR" },
	{ GPIO_FN_FSIAOSLD,		0,	"FSIAOSLD" },
	{ GPIO_FN_FSIAISLD_PORT5,	0,	"FSIAISLD" },
};

static void fsi_wm8978_platform_init(struct snd_soc_pcm_runtime *rtd)
{
	struct snd_soc_dai *codec = rtd->codec_dai;
	snd_soc_dapm_force_enable_pin(&codec->codec->dapm, "Mic Bias");
}

static struct asoc_simple_dai_init_info __maybe_unused fsi_wm8978_init_info = {
	.fmt		= SND_SOC_DAIFMT_I2S,
	.codec_daifmt	= SND_SOC_DAIFMT_CBM_CFM | SND_SOC_DAIFMT_NB_NF,
	.cpu_daifmt	= SND_SOC_DAIFMT_CBS_CFS,
	.sysclk_id	= 0, /* WM8978_PLL */
	.sysclk		= 12288000,
};

static struct asoc_simple_card_info __maybe_unused fsi_wm8978_info = {
	.name		= "wm8978",
	.card		= "FSI2A-WM8978",
	.cpu_dai	= "fsia-dai",
	.codec		= "wm8978.0-001a",
	.platform	= "sh_fsi2",
	.codec_dai	= "wm8978-hifi",
	.init		= &fsi_wm8978_init_info,
	.platform_widget_init = fsi_wm8978_platform_init,
};

static struct platform_device __maybe_unused fsi_wm8978_device = {
	.name	= "asoc-simple-card",
	.id	= 0,
	.dev	= {
		.platform_data	= &fsi_wm8978_info,
	},
};

static struct asoc_simple_dai_init_info fsi2_hdmi_init_info = {
	.cpu_daifmt	= SND_SOC_DAIFMT_CBM_CFM,
};

static struct asoc_simple_card_info fsi2_hdmi_info = {
	.name		= "HDMI",
	.card		= "FSI2B-HDMI",
	.cpu_dai	= "fsib-dai",
	.codec		= "sh-mobile-hdmi",
	.platform	= "sh_fsi2",
	.codec_dai	= "sh_mobile_hdmi-hifi",
	.init		= &fsi2_hdmi_init_info,
};

static struct platform_device fsi_hdmi_device = {
	.name	= "asoc-simple-card",
	.id	= 1,
	.dev	= {
		.platform_data	= &fsi2_hdmi_info,
	},
};

static void a840_audio_init(void)
{
	if (gpio_request(GPIO_FN_FSIBCK, NULL)) {
		pr_err("FSIBCK is already reserved.\n");
		return;
	}
	gpio_request_pulldown(portcr(11));

	/* AUDIO_CLK_EN */
	gpio_request(GPIO_PORT19, NULL);
	gpio_direction_output(GPIO_PORT19, 1);

	platform_device_register(&fsi_device);
	platform_device_register(&fsi_hdmi_device);

	if (con7eb_detected &&
	    IS_ENABLED(CONFIG_ARMADILLO840_FSIA_SLAVE_DEVICE_WM8978)) {
		if (gpio_request_array_simple(fsia_gpios,
					      ARRAY_SIZE(fsia_gpios))) {
			pr_err("Any FSI-A resources are already reserved.\n");
			return;
		}

		platform_device_register(&fsi_wm8978_device);
	}
}

/* PWM */
static struct resource pwm_resources[] = {
	[0] = {
		.start = 0xe6600000,
		.end = 0xe66000ff,
		.flags = IORESOURCE_MEM,
	},
};

static struct rmob_tpu_pwm_platform_data pwm_device_data = {
	.channels[2] = {
		.polarity = 0,
		.pin_gpio = GPIO_PORT202,
		.pin_gpio_fn = GPIO_FN_TPU0TO2_PORT202,
	}
};

static struct platform_device pwm_device = {
	.name = "rmob_tpu_pwm",
	.id = 0,
	.dev = {
		.platform_data = &pwm_device_data,
	},
	.num_resources = ARRAY_SIZE(pwm_resources),
	.resource = pwm_resources,
};

static struct platform_pwm_backlight_data pwm_backlight_data = {
	.pwm_id = RMOB_TPU_PWM_ID(0, 2),
	.lth_brightness = 50,
	.max_brightness = 255,
	.dft_brightness = 255,
	.pwm_period_ns = 33333, /* 30kHz */
};

static struct platform_device pwm_backlight_device = {
	.name = "pwm-backlight",
	.dev = {
		.platform_data = &pwm_backlight_data,
	},
};

/* LCDC0 */
static struct fb_videomode lcdc0_mode = {
#if IS_ENABLED(CONFIG_ARMADILLO840_LCDC0_DEVICE_AM800480)
	.name		= "AMPIRE/AM-800480",
	.xres		= 800,
	.yres		= 480,
	.pixclock	= 33670,
	.left_margin	= 88,
	.right_margin	= 40,
	.hsync_len	= 128,
	.upper_margin	= 20,
	.lower_margin	= 5,
	.vsync_len	= 5,
	.refresh	= 55,
	.sync		= 0,
	.vmode		= FB_VMODE_NONINTERLACED,
	.flag		= FB_MODE_IS_DETAILED,
#elif IS_ENABLED(CONFIG_ARMADILLO840_LCDC0_DEVICE_SCF0500)
	.name		= "DATAIMAGE/SCF0500",
	.xres		= 800,
	.yres		= 480,
	.pixclock	= 26936,
	.left_margin	= 88,
	.right_margin	= 40,
	.hsync_len	= 128,
	.upper_margin	= 20,
	.lower_margin	= 5,
	.vsync_len	= 5,
	.refresh	= 69,
	.sync		= 0,
	.vmode		= FB_VMODE_NONINTERLACED,
	.flag		= FB_MODE_IS_DETAILED,
#endif
};

static const struct sh_videomode_hint lcdc0_mode_hint = {
	.clock_source	= LCDC_CLK_BUS,
	.device_type	= LCDC_DEV_LCD,
	.interface_type	= RGB24,
};

static struct sh_mobile_lcdc_info lcdc0_info = {
	.clock_source	= LCDC_CLK_BUS,
	.ch[0] = {
		.chan		= LCDC_CHAN_MAINLCD,
		.fourcc		= V4L2_PIX_FMT_BGR32,
		.interface_type	= RGB24,
		.clock_divider	= 5,
		.flags		= 0,
		.lcd_modes	= &lcdc0_mode,
		.lcd_mode_hints	= &lcdc0_mode_hint,
		.num_modes	= 1,
		.panel_cfg = {
			.width	= 111,
			.height = 68,
		},
	},
};

static struct resource lcdc0_resources[] = {
	[0] = {
		.name	= "LCD0",
		.start	= 0xfe940000,
		.end	= 0xfe943fff,
		.flags	= IORESOURCE_MEM,
	},
	[1] = {
		.start	= gic_spi(177),
		.flags	= IORESOURCE_IRQ,
	},
};

static struct platform_device lcdc0_device = {
	.name		= "sh_mobile_lcdc_fb",
	.num_resources	= ARRAY_SIZE(lcdc0_resources),
	.resource	= lcdc0_resources,
	.id		= 0,
	.dev	= {
		.platform_data	= &lcdc0_info,
		.coherent_dma_mask = ~0,
	},
};

static struct gpio __maybe_unused lcdc0_gpios[] = {
	{ GPIO_FN_LCD0_D0,		0,	"LCD0_D0" },
	{ GPIO_FN_LCD0_D1,		0,	"LCD0_D1" },
	{ GPIO_FN_LCD0_D2,		0,	"LCD0_D2" },
	{ GPIO_FN_LCD0_D3,		0,	"LCD0_D3" },
	{ GPIO_FN_LCD0_D4,		0,	"LCD0_D4" },
	{ GPIO_FN_LCD0_D5,		0,	"LCD0_D5" },
	{ GPIO_FN_LCD0_D6,		0,	"LCD0_D6" },
	{ GPIO_FN_LCD0_D7,		0,	"LCD0_D7" },
	{ GPIO_FN_LCD0_D8,		0,	"LCD0_D8" },
	{ GPIO_FN_LCD0_D9,		0,	"LCD0_D9" },
	{ GPIO_FN_LCD0_D10,		0,	"LCD0_D10" },
	{ GPIO_FN_LCD0_D11,		0,	"LCD0_D11" },
	{ GPIO_FN_LCD0_D12,		0,	"LCD0_D12" },
	{ GPIO_FN_LCD0_D13,		0,	"LCD0_D13" },
	{ GPIO_FN_LCD0_D14,		0,	"LCD0_D14" },
	{ GPIO_FN_LCD0_D15,		0,	"LCD0_D15" },
	{ GPIO_FN_LCD0_D16,		0,	"LCD0_D16" },
	{ GPIO_FN_LCD0_D17,		0,	"LCD0_D17" },
	{ GPIO_FN_LCD0_DCK,		0,	"LCD0_DCK" },
	{ GPIO_FN_LCD0_VSYN,		0,	"LCD0_VSYN" },
	{ GPIO_FN_LCD0_HSYN,		0,	"LCD0_HSYN" },
	{ GPIO_FN_LCD0_DISP,		0,	"LCD0_DISP" },
	{ GPIO_PORT61,			0,	"LCD0_DON" },
};

static struct gpio __maybe_unused lcdc0_24bit_gpios[] = {
	{ GPIO_FN_LCD0_D18_PORT40,	0,	"LCD0_D18" },
	{ GPIO_FN_LCD0_D19_PORT4,	0,	"LCD0_D19" },
	{ GPIO_FN_LCD0_D20_PORT3,	0,	"LCD0_D20" },
	{ GPIO_FN_LCD0_D21_PORT2,	0,	"LCD0_D21" },
	{ GPIO_FN_LCD0_D22_PORT0,	0,	"LCD0_D22" },
	{ GPIO_FN_LCD0_D23_PORT1,	0,	"LCD0_D23" },
};

static struct gpio __maybe_unused lcdc0_lclk_gpios[] = {
	{ GPIO_FN_LCD0_LCLK_PORT165,	0,	"LCD0_LCLK" },
};

static struct gpio __maybe_unused ts_st1232_gpios[] = {
	{ GPIO_FN_IRQ4_PORT172,		0,	NULL },
	{ GPIO_PORT172,			0,	"TP_INT" },
	{ GPIO_PORT173,			0,	"TP_RST" },
};

static void a840_lcdc0_init(void)
{
	int ret;

	if (!(con7eb_detected && IS_ENABLED(CONFIG_ARMADILLO840_LCDC0)))
		return;

	ret = gpio_request_array_simple(lcdc0_gpios, ARRAY_SIZE(lcdc0_gpios));
	if (!ret && IS_ENABLED(CONFIG_ARMADILLO840_LCDC0_HAVE_D18_D23))
		ret = gpio_request_array_simple(lcdc0_24bit_gpios,
						ARRAY_SIZE(lcdc0_24bit_gpios));
	if (!ret && IS_ENABLED(CONFIG_ARMADILLO840_LCDC0_HAVE_LCLK0_PORT165))
		ret = gpio_request_array_simple(lcdc0_lclk_gpios,
						ARRAY_SIZE(lcdc0_lclk_gpios));
	if (ret) {
		pr_err("Any LCDC0 resources are already reserved.\n");
		return;
	}

	/* SCF0500: dotclk: 37.125MHz (M3:148.5MHz / 4) */
	if (IS_ENABLED(CONFIG_ARMADILLO840_LCDC0_DEVICE_SCF0500))
		lcdc0_info.ch[0].clock_divider = 4;

	gpio_request(GPIO_FN_LCDC0_SELECT, NULL);
	/* unused Display-ON(DON). force ON */
	gpio_direction_output(GPIO_PORT61, 1);

	platform_device_register(&lcdc0_device);

	/* touch-screen */
	if (IS_ENABLED(CONFIG_ARMADILLO840_I2C0_DEVICE_ST1232)) {
		if (gpio_request_array_simple(ts_st1232_gpios,
					      ARRAY_SIZE(ts_st1232_gpios)))
			pr_err("Any ST1232 resources are already reserved.\n");
		/* reset of st1232 is controlled by PORT173 */
		gpio_direction_output(GPIO_PORT173, 0);
		udelay(1);
		gpio_direction_output(GPIO_PORT173, 1);
		gpio_direction_input(GPIO_PORT172);
	}

	/* BL_SHDN_B of TB62752 LED driver */
	if (IS_ENABLED(CONFIG_ARMADILLO840_LED_DRIVER_TB62752)) {
		gpio_request(GPIO_PORT164, NULL);
		gpio_direction_output(GPIO_PORT164, 1);
	}

	/* pwm for backlight */
	if (IS_ENABLED(CONFIG_ARMADILLO840_TPU0TO2_PORT202)) {
		gpio_reserve(GPIO_PORT202);

		platform_device_register(&pwm_device);
	}
	/* backlight */
	if (IS_ENABLED(CONFIG_ARMADILLO840_TPU0TO2_PORT202_DEVICE_PWM_BACKLIGHT))
		platform_device_register(&pwm_backlight_device);
}

/* LCDC1/HDMI/SDENC */
#define SDENC_PDCR	0xe6058148

static struct sh_mobile_sdenc_info __maybe_unused sdenc_info = {
};

static struct resource __maybe_unused sdenc_resources[] = {
	[0] = {
		.name	= "SDENC",
		.start	= 0xffcb0000,
		.end	= 0xffcb03ff,
		.flags	= IORESOURCE_MEM,
	},
};

static struct platform_device __maybe_unused sdenc_device = {
	.name		= "sh-mobile-sdenc",
	.id		= 0,
	.num_resources	= ARRAY_SIZE(sdenc_resources),
	.resource	= sdenc_resources,
	.dev		= {
		.platform_data	= &sdenc_info,
	},
};

static struct sh_mobile_hdmi_info hdmi_info = {
	.flags		= HDMI_OUTPUT_PUSH_PULL |
			  HDMI_OUTPUT_POLARITY_HI |
			  HDMI_32BIT_REG |
			  HDMI_HAS_HTOP1 |
			  HDMI_SND_SRC_SPDIF,
};

static struct resource hdmi_resources[] = {
	[0] = {
		.name	= "HDMI",
		.start	= 0xe6be0000,
		.end	= 0xe6be03ff,
		.flags	= IORESOURCE_MEM,
	},
	[1] = {
		.start	= gic_spi(131),
		.flags	= IORESOURCE_IRQ,
	},
	[2] = {
		.name	= "HDMI emma3pf",
		.start	= 0xe6be4000,
		.end	= 0xe6be43ff,
		.flags	= IORESOURCE_MEM,
	},
};

static struct platform_device hdmi_device = {
	.name		= "sh-mobile-hdmi",
	.num_resources	= ARRAY_SIZE(hdmi_resources),
	.resource	= hdmi_resources,
	.id             = -1,
	.dev	= {
		.platform_data	= &hdmi_info,
	},
};

enum {
#if IS_ENABLED(CONFIG_ARMADILLO840_SDENC)
	LCDC1_VMODE_NTSC,
#endif
	LCDC1_VMODE_1080I_60,
	LCDC1_VMODE_1080P_60,
	LCDC1_VMODE_1080P_24,
	LCDC1_VMODE_720P_60,
	LCDC1_VMODE_480P_59,
};

static const struct fb_videomode lcdc1_modes[] = {
#if IS_ENABLED(CONFIG_ARMADILLO840_SDENC)
	[LCDC1_VMODE_NTSC] = {
		.name		= "NTSC",
		.xres		= 720,
		.yres		= 480,
		.pixclock	= 37037,
		.left_margin	= 60,
		.right_margin	= 16,
		.hsync_len	= 62,
		.upper_margin	= 30,
		.lower_margin	= 9,
		.vsync_len	= 6,
		.refresh	= 60,
		.sync		= 0,
		.vmode		= FB_VMODE_INTERLACED,
		.flag		= FB_MODE_IS_DETAILED,
	},
#endif
	[LCDC1_VMODE_1080I_60] = {
		.name		= "1920x1080i@60",
		.xres		= 1920,
		.yres		= 1080,
		.pixclock	= 13468,
		.left_margin	= 148,
		.right_margin	= 88,
		.hsync_len	= 44,
		.upper_margin	= 30,
		.lower_margin	= 4,
		.vsync_len	= 10,
		.refresh	= 60,
		.sync		= FB_SYNC_VERT_HIGH_ACT | FB_SYNC_HOR_HIGH_ACT,
		.vmode		= FB_VMODE_INTERLACED,
		.flag		= FB_MODE_IS_DETAILED,
	},
	[LCDC1_VMODE_1080P_60] = {
		.name		= "1920x1080@60",
		.xres		= 1920,
		.yres		= 1080,
		.pixclock	= 6734,
		.left_margin	= 148,
		.right_margin	= 88,
		.hsync_len	= 44,
		.upper_margin	= 36,
		.lower_margin	= 4,
		.vsync_len	= 5,
		.refresh	= 60,
		.sync		= FB_SYNC_VERT_HIGH_ACT | FB_SYNC_HOR_HIGH_ACT,
		.vmode		= FB_VMODE_NONINTERLACED,
		.flag		= FB_MODE_IS_DETAILED,
	},
	[LCDC1_VMODE_1080P_24] = {
		.name		= "1920x1080@24",
		.xres		= 1920,
		.yres		= 1080,
		.pixclock	= 13468,
		.left_margin	= 148,
		.right_margin	= 638,
		.hsync_len	= 44,
		.upper_margin	= 36,
		.lower_margin	= 4,
		.vsync_len	= 5,
		.refresh	= 24,
		.sync		= FB_SYNC_VERT_HIGH_ACT | FB_SYNC_HOR_HIGH_ACT,
		.vmode		= FB_VMODE_NONINTERLACED,
		.flag		= FB_MODE_IS_DETAILED,
	},
	[LCDC1_VMODE_720P_60] = {
		.name		= "1280x720@60",
		.xres		= 1280,
		.yres		= 720,
		.pixclock	= 13468,
		.left_margin	= 220,
		.right_margin	= 110,
		.hsync_len	= 40,
		.upper_margin	= 20,
		.lower_margin	= 5,
		.vsync_len	= 5,
		.refresh	= 60,
		.sync		= FB_SYNC_VERT_HIGH_ACT | FB_SYNC_HOR_HIGH_ACT,
		.vmode		= FB_VMODE_NONINTERLACED,
		.flag		= FB_MODE_IS_DETAILED,
	},
	[LCDC1_VMODE_480P_59] = {
		.name		= "720x480@59",
		.xres		= 720,
		.yres		= 480,
		.pixclock	= 37037,
		.left_margin	= 60,
		.right_margin	= 16,
		.hsync_len	= 62,
		.upper_margin	= 30,
		.lower_margin	= 9,
		.vsync_len	= 6,
		.refresh	= 59,
		.sync		= 0,
		.vmode		= FB_VMODE_NONINTERLACED,
		.flag		= FB_MODE_IS_DETAILED,
	},
};

#define LCDC1_DEFAULT_CLKSRC	LCDC_CLK_PERIPHERAL
#define LCDC1_DEFAULT_DEVTYPE	LCDC_DEV_HDMI
#define LCDC1_DEFAULT_IFTYPE	RGB24
#define LCDC1_DEFAULT_TXDEV	&hdmi_device

static const struct sh_videomode_hint lcdc1_mode_hints[] = {
#if IS_ENABLED(CONFIG_ARMADILLO840_SDENC)
	[LCDC1_VMODE_NTSC] = {
		.clock_source		= LCDC_CLK_EXTERNAL2,
		.device_type		= LCDC_DEV_SDENC,
		.interface_type		= YUV422D,
		.tx_dev			= &sdenc_device,
	},
#endif
	[LCDC1_VMODE_1080I_60] = {
		.clock_source		= LCDC1_DEFAULT_CLKSRC,
		.device_type		= LCDC1_DEFAULT_DEVTYPE,
		.interface_type		= LCDC1_DEFAULT_IFTYPE,
		.tx_dev			= LCDC1_DEFAULT_TXDEV,
	},

	[LCDC1_VMODE_1080P_60] = {
		.clock_source		= LCDC1_DEFAULT_CLKSRC,
		.device_type		= LCDC1_DEFAULT_DEVTYPE,
		.interface_type		= LCDC1_DEFAULT_IFTYPE,
		.tx_dev			= LCDC1_DEFAULT_TXDEV,
	},
	[LCDC1_VMODE_1080P_24] = {
		.clock_source		= LCDC1_DEFAULT_CLKSRC,
		.device_type		= LCDC1_DEFAULT_DEVTYPE,
		.interface_type		= LCDC1_DEFAULT_IFTYPE,
		.tx_dev			= LCDC1_DEFAULT_TXDEV,
	},
	[LCDC1_VMODE_720P_60] = {
		.clock_source		= LCDC1_DEFAULT_CLKSRC,
		.device_type		= LCDC1_DEFAULT_DEVTYPE,
		.interface_type		= LCDC1_DEFAULT_IFTYPE,
		.tx_dev			= LCDC1_DEFAULT_TXDEV,
	},
	[LCDC1_VMODE_480P_59] = {
		.clock_source		= LCDC1_DEFAULT_CLKSRC,
		.device_type		= LCDC1_DEFAULT_DEVTYPE,
		.interface_type		= LCDC1_DEFAULT_IFTYPE,
		.tx_dev			= LCDC1_DEFAULT_TXDEV,
	},
};

static struct sh_mobile_lcdc_info lcdc1_info = {
	.clock_source	= LCDC_CLK_PERIPHERAL, /* HDMI clock */
	.ch[0] = {
		.chan			= LCDC_CHAN_MAINLCD,
		.fourcc			= V4L2_PIX_FMT_BGR32,
		.interface_type		= RGB24,
		.clock_divider		= 1,
		.flags			= LCDC_FLAGS_DWPOL,
		.lcd_modes		= lcdc1_modes,
		.lcd_mode_hints		= lcdc1_mode_hints,
		.num_modes		= ARRAY_SIZE(lcdc1_modes),
		.tx_dev			= &hdmi_device,
	},
};

static struct resource lcdc1_resources[] = {
	[0] = {
		.name	= "LCDC1",
		.start	= 0xfe944000,
		.end	= 0xfe948000 - 1,
		.flags	= IORESOURCE_MEM,
	},
	[1] = {
		.start	= gic_spi(178),
		.flags	= IORESOURCE_IRQ,
	},
};

static struct platform_device lcdc1_device = {
	.name		= "sh_mobile_lcdc_fb",
	.num_resources	= ARRAY_SIZE(lcdc1_resources),
	.resource	= lcdc1_resources,
	.id		= 1,
	.dev	= {
		.platform_data	= &lcdc1_info,
		.coherent_dma_mask = ~0,
	},
};

static struct gpio hdmi_gpios[] = {
	{ GPIO_FN_HDMI_HPD,		0,	"HPD" },
	{ GPIO_FN_HDMI_CEC,		0,	"CEC" },
};

static void a840_lcdc1_init(void)
{
	lcdc1_info.ch[0].caps = LCDC_CAP_HDMI;

	if (IS_ENABLED(CONFIG_ARMADILLO840_SDENC)) {
		if (con7eb_detected) {
			lcdc1_info.ch[0].caps |= LCDC_CAP_SDENC;

			__raw_writew(0x0001, SDENC_PDCR);
			gpio_request(GPIO_FN_SDENC_DV_CLKI, NULL);
			if (IS_ENABLED(CONFIG_ARMADILLO840_SDENC_HAVE_PSAVE_PORT165)) {
				gpio_reserve(GPIO_PORT165);
				sdenc_info.gpio_pwrsave = GPIO_PORT165;
				sdenc_info.gpio_pwrsave_active_low = 1;
			}

			platform_device_register(&sdenc_device);
		} else {
			lcdc1_info.ch[0].lcd_modes = &lcdc1_modes[1];
			lcdc1_info.ch[0].lcd_mode_hints = \
				&lcdc1_mode_hints[1];
			lcdc1_info.ch[0].num_modes--;
		}
	}

	if (gpio_request_array_simple(hdmi_gpios, ARRAY_SIZE(hdmi_gpios))) {
		pr_err("Any HDMI resources are already reserved.\n");
		return;
	}

	platform_device_register(&hdmi_device);
	platform_device_register(&lcdc1_device);
}

/* CEC */
static struct resource cec_resources[] = {
	[0] = {
		.start	= 0xe6bf0000,
		.end	= 0xe6bf0020 - 1,
		.flags	= IORESOURCE_MEM,
	},
	[1] = {
		.start	= gic_spi(3),
		.flags	= IORESOURCE_IRQ,
	},
	[2] = {
		.start	= gic_spi(4),
		.flags	= IORESOURCE_IRQ,
	},
};

#define __a840_lcdc_init(n) a840_lcdc##n##_init()
#define _a840_lcdc_init(n) __a840_lcdc_init(n)

static void a840_lcdc_init(void)
{
	_a840_lcdc_init(CONFIG_ARMADILLO840_PRIMARY_FB_LCDC);
	_a840_lcdc_init(CONFIG_ARMADILLO840_SECONDARY_FB_LCDC);
}

static struct platform_device cec_device = {
	.name		= "r8a7740_cec",
	.id		= 0,
	.resource	= cec_resources,
	.num_resources	= ARRAY_SIZE(cec_resources),
};

/* CAMERA */
static int __maybe_unused camera_clock_enable(struct device *dev,
					      const char *clkname,
					      unsigned long rate)
{
	struct clk *mclk;

	mclk = clk_get(dev, clkname);
	if (IS_ERR(mclk)) {
		dev_err(dev, "can't get %s clock\n", clkname);
		return -EINVAL;
	}
	clk_set_rate(mclk, clk_round_rate(mclk, rate));
	clk_enable(mclk);
	clk_put(mclk);

	return 0;
}

static int __maybe_unused camera_clock_disable(struct device *dev,
					       const char *clkname)
{
	struct clk *mclk;

	mclk = clk_get(NULL, clkname);
	if (IS_ERR(mclk)) {
		dev_err(dev, "can't get %s clock\n", clkname);
		return -EINVAL;
	}
	clk_disable(mclk);
	clk_put(mclk);

	return 0;
}

static int __maybe_unused ov7725_reset(struct device *dev)
{
	gpio_direction_output(GPIO_PORT195, 0);
	mdelay(1);
	gpio_direction_output(GPIO_PORT195, 1);

	return 0;
}

static int __maybe_unused ov7725_power(struct device *dev, int mode)
{
	struct platform_device *pdev = to_platform_device(dev);
	const char *clkname = pdev->id == 0 ? "video1" : "video2";

	if (mode) {
		gpio_direction_output(GPIO_PORT196, 0);
		return camera_clock_enable(dev, clkname, 11880000);
	} else {
		gpio_direction_output(GPIO_PORT196, 1);
		return camera_clock_disable(dev, clkname);
	}
}

static struct ov772x_camera_info __maybe_unused ov7725_info = {
	.edgectrl	= OV772X_AUTO_EDGECTRL(0xf, 0),
};

static struct i2c_board_info __maybe_unused i2c_ov7725_device = {
	I2C_BOARD_INFO("ov772x", 0x21),
};

/* CONFIG_ARMADILLO840_CEU1_DEVICE_KBCRIC01VG */
static struct soc_camera_link __maybe_unused camera_link_kbcric01vg_con7 = {
	.bus_id		= 1,
	.module_name	= "ov772x",
	.i2c_adapter_id	= 3,
	.board_info	= &i2c_ov7725_device,
	.power		= ov7725_power,
	.reset		= ov7725_reset,
	.priv		= &ov7725_info,
};

/* CONFIG_ARMADILLO840_CEU0_DEVICE_KBCRIC01VG */
static struct soc_camera_link __maybe_unused camera_link_kbcric01vg_con8 = {
	.bus_id		= 0,
	.module_name	= "ov772x",
	.i2c_adapter_id	= 1,
	.board_info	= &i2c_ov7725_device,
	.power		= ov7725_power,
	.reset		= ov7725_reset,
	.priv		= &ov7725_info,
};

/* CEU0 */
static struct platform_device __maybe_unused ceu0_camera_device = {
	.name	= "soc-camera-pdrv",
	.id	= 0,
	.dev	= {
		.platform_data = NULL,
	},
};

static struct sh_mobile_ceu_info __maybe_unused ceu0_info = {
	.flags	= 0,
};

static struct resource __maybe_unused ceu0_resources[] = {
	[0] = {
		.name	= "CEU",
		.start	= 0xfe910000,
		.end	= 0xfe91009f,
		.flags	= IORESOURCE_MEM,
	},
	[1] = {
		.start  = gic_spi(160),
		.flags  = IORESOURCE_IRQ,
	},
	[2] = {
		/* place holder for contiguous memory */
	},
};

static struct platform_device __maybe_unused ceu0_device = {
	.name		= "sh_mobile_ceu",
	.id		= 0,
	.num_resources	= ARRAY_SIZE(ceu0_resources),
	.resource	= ceu0_resources,
	.dev		= {
		.platform_data		= &ceu0_info,
		.coherent_dma_mask	= 0xffffffff,
	},
};

/* CEU1 */
static struct platform_device __maybe_unused ceu1_camera_device = {
	.name	= "soc-camera-pdrv",
	.id	= 1,
	.dev	= {
		.platform_data = NULL,
	},
};

static struct sh_mobile_ceu_info __maybe_unused ceu1_info = {
	.flags	= SH_CEU_FLAG_USE_8BIT_BUS | SH_CEU_FLAG_LOWER_8BIT,
};

static struct resource __maybe_unused ceu1_resources[] = {
	[0] = {
		.name	= "CEU",
		.start	= 0xfe914000,
		.end	= 0xfe91409f,
		.flags	= IORESOURCE_MEM,
	},
	[1] = {
		.start  = gic_spi(159),
		.flags  = IORESOURCE_IRQ,
	},
	[2] = {
		/* place holder for contiguous memory */
	},
};

static struct platform_device __maybe_unused ceu1_device = {
	.name		= "sh_mobile_ceu",
	.id		= 1,
	.num_resources	= ARRAY_SIZE(ceu1_resources),
	.resource	= ceu1_resources,
	.dev		= {
		.platform_data		= &ceu1_info,
		.coherent_dma_mask	= 0xffffffff,
	},
};

static struct gpio __maybe_unused ceu0_upper_8bits_gpios[] = {
	{ GPIO_FN_VIO0_D15_PORT24,	0,	"CEU0_D15" },
	{ GPIO_FN_VIO0_D14_PORT25,	0,	"CEU0_D14" },
	{ GPIO_FN_VIO0_D13_PORT26,	0,	"CEU0_D13" },
	{ GPIO_FN_VIO0_D12,		0,	"CEU0_D12" },
	{ GPIO_FN_VIO0_D11,		0,	"CEU0_D11" },
	{ GPIO_FN_VIO0_D10,		0,	"CEU0_D10" },
	{ GPIO_FN_VIO0_D9,		0,	"CEU0_D9" },
	{ GPIO_FN_VIO0_D8,		0,	"CEU0_D8" },
};

static struct gpio __maybe_unused ceu0_lower_8bits_gpios[] = {
	{ GPIO_FN_VIO0_D7,		0,	"CEU0_D7" },
	{ GPIO_FN_VIO0_D6,		0,	"CEU0_D6" },
	{ GPIO_FN_VIO0_D5,		0,	"CEU0_D5" },
	{ GPIO_FN_VIO0_D4,		0,	"CEU0_D4" },
	{ GPIO_FN_VIO0_D3,		0,	"CEU0_D3" },
	{ GPIO_FN_VIO0_D2,		0,	"CEU0_D2" },
	{ GPIO_FN_VIO0_D1,		0,	"CEU0_D1" },
	{ GPIO_FN_VIO0_D0,		0,	"CEU0_D0" },
};

static struct gpio __maybe_unused ceu0_common_gpios[] = {
	{ GPIO_FN_VIO0_CLK,		0,	"CEU0_CLK" },
	{ GPIO_FN_VIO0_HD,		0,	"CEU0_HD" },
	{ GPIO_FN_VIO0_VD,		0,	"CEU0_VD" },
	{ GPIO_FN_VIO0_FIELD,		0,	"CEU0_FIELD" },
	{ GPIO_FN_VIO_CKO,		0,	"CEU0_CKO" },
};

static struct gpio __maybe_unused ceu1_gpios[] = {
	{ GPIO_FN_VIO1_D7,		0,	"CEU1_D7" },
	{ GPIO_FN_VIO1_D6,		0,	"CEU1_D6" },
	{ GPIO_FN_VIO1_D5,		0,	"CEU1_D5" },
	{ GPIO_FN_VIO1_D4,		0,	"CEU1_D4" },
	{ GPIO_FN_VIO1_D3,		0,	"CEU1_D3" },
	{ GPIO_FN_VIO1_D2,		0,	"CEU1_D2" },
	{ GPIO_FN_VIO1_D1,		0,	"CEU1_D1" },
	{ GPIO_FN_VIO1_D0,		0,	"CEU1_D0" },
	{ GPIO_FN_VIO1_CLK,		0,	"CEU1_CLK" },
	{ GPIO_FN_VIO1_HD,		0,	"CEU1_HD" },
	{ GPIO_FN_VIO1_VD,		0,	"CEU1_VD" },
	{ GPIO_FN_VIO1_FIELD,		0,	"CEU1_FIELD" },
	{ GPIO_FN_VIO_CKO_1,		0,	"CEU1_CKO_1" },
};

static void a840_camera_init(void)
{
	int ceu0_status = 0;
	int ceu1_status = 0;

	if (IS_ENABLED(CONFIG_ARMADILLO840_CEU0_LOWER)) {
		ceu0_status = gpio_request_array_simple(ceu0_lower_8bits_gpios,
					ARRAY_SIZE(ceu0_lower_8bits_gpios));
		if (ceu0_status)
			pr_err("Any CEU0 resources (D[7-0]) "
			       "are already reserved.\n");
		else
			ceu0_status = 1;
	}

	if (ceu0_status >= 0 &&
	    IS_ENABLED(CONFIG_ARMADILLO840_CEU0_UPPER)) {
		ceu0_status = gpio_request_array_simple(ceu0_upper_8bits_gpios,
					ARRAY_SIZE(ceu0_upper_8bits_gpios));
		if (ceu0_status)
			pr_err("Any CEU0 resources (D[15-8]) "
			       "are already reserved.\n");
		else
			ceu0_status = 1;
	}

	if (ceu0_status > 0) {
		ceu0_status = gpio_request_array_simple(ceu0_common_gpios,
					ARRAY_SIZE(ceu0_common_gpios));
		if (ceu0_status)
			pr_err("Any CEU0 resources (CLKs and SYNCs) "
			       "are already reserved.\n");
		else
			ceu0_status = 1;
	}

	if (IS_ENABLED(CONFIG_ARMADILLO840_CEU1)) {
		ceu1_status = gpio_request_array_simple(ceu1_gpios,
					ARRAY_SIZE(ceu1_gpios));
		if (ceu1_status)
			pr_err("Any CEU1 resources are already reserved.\n");
		else
			ceu1_status = 1;
	}

	if (IS_ENABLED(CONFIG_ARMADILLO840_CEU0_DEVICE_KBCRIC01VG) ||
	    IS_ENABLED(CONFIG_ARMADILLO840_CEU1_DEVICE_KBCRIC01VG)) {
		gpio_request(GPIO_PORT195, NULL); /* RSTB */
		gpio_request(GPIO_PORT196, NULL); /* PWDN */
		gpio_direction_output(GPIO_PORT195, 1);
		gpio_direction_output(GPIO_PORT196, 0);
	}
	if (IS_ENABLED(CONFIG_ARMADILLO840_CEU0_DEVICE_KBCRIC01VG)) {
		ceu0_info.flags = SH_CEU_FLAG_USE_8BIT_BUS; /* upper 8bit */
		ceu0_camera_device.dev.platform_data = &camera_link_kbcric01vg_con8;
	}
	if (IS_ENABLED(CONFIG_ARMADILLO840_CEU1_DEVICE_KBCRIC01VG)) {
		ceu1_camera_device.dev.platform_data = &camera_link_kbcric01vg_con7;
	}

	if (ceu0_status > 0) {
		platform_device_register(&ceu0_device);
		platform_device_register(&ceu0_camera_device);
	}
	if (ceu1_status > 0) {
		platform_device_register(&ceu1_device);
		platform_device_register(&ceu1_camera_device);
	}
}

/* Flash memory: MTD-physmap */
#define MTD_PART(_name, _offset, _size, ...) \
	{ .name = _name, .offset = _offset, .size = _size, __VA_ARGS__ }

static struct mtd_partition flash_partitions[] = {
	MTD_PART("bootloader",	0x00000000,		SZ_256K,
		 .mask_flags = MTD_WRITEABLE),
	MTD_PART("config",	MTDPART_OFS_APPEND,	SZ_256K),
	MTD_PART("license",	MTDPART_OFS_APPEND,	SZ_256K,
		 .mask_flags = MTD_WRITEABLE),
	MTD_PART("firmware",	MTDPART_OFS_APPEND,	SZ_4M,
		 .mask_flags = MTD_WRITEABLE),
	MTD_PART("kernel",	MTDPART_OFS_APPEND,	SZ_4M),
	MTD_PART("userland",	MTDPART_OFS_APPEND,	MTDPART_SIZ_FULL),
};

static struct physmap_flash_data flash_data = {
	.width		= 2,
	.parts		= flash_partitions,
	.nr_parts	= ARRAY_SIZE(flash_partitions),
};

static struct resource flash_resources[] = {
	[0] = {
		.start	= 0x04000000,
		.end	= 0x0c000000 - 1,
		.flags	= IORESOURCE_MEM,
	}
};

static struct platform_device flash_device = {
	.name		= "physmap-flash",
	.dev		= {
		.platform_data	= &flash_data,
	},
	.num_resources	= ARRAY_SIZE(flash_resources),
	.resource	= flash_resources,
};

/* LED */
#define GPIO_LED(_name, _default_trigger, _gpio, _active_low)	\
{								\
	.name			= _name,			\
	.default_trigger	= _default_trigger,		\
	.gpio			= _gpio,			\
	.active_low		= _active_low,			\
}

static struct gpio_led gpio_leds_pins[] = {
	GPIO_LED("LED1", "none", GPIO_PORT107, 0),
	GPIO_LED("LED2", "none", GPIO_PORT108, 0),
#if IS_ENABLED(CONFIG_ARMADILLO840_LED_EXT1_PORT102)
	GPIO_LED("EXT1", "none", GPIO_PORT102, 0),
#endif
#if IS_ENABLED(CONFIG_ARMADILLO840_LED_EXT2_PORT59)
	GPIO_LED("EXT2", "none", GPIO_PORT59, 0),
#endif
#if IS_ENABLED(CONFIG_ARMADILLO840_LED_EXT3_PORT60)
	GPIO_LED("EXT3", "none", GPIO_PORT60, 0),
#endif
#if IS_ENABLED(CONFIG_ARMADILLO840_LED_EXT4_PORT10)
	GPIO_LED("EXT4", "none", GPIO_PORT10, 0),
#endif
#if IS_ENABLED(CONFIG_ARMADILLO840_LED_EXT5_PORT8)
	GPIO_LED("EXT5", "none", GPIO_PORT8, 0),
#endif
#if IS_ENABLED(CONFIG_ARMADILLO840_LED_EXT6_PORT7)
	GPIO_LED("EXT6", "none", GPIO_PORT7, 0),
#endif
};

static struct gpio_led_platform_data gpio_leds_data = {
	.leds		= gpio_leds_pins,
	.num_leds	= ARRAY_SIZE(gpio_leds_pins),
};

static struct platform_device gpio_leds_device = {
	.name		= "leds-gpio",
	.id		= -1,
	.dev		= {
		.platform_data	= &gpio_leds_data,
	},
};

static void a840_led_init(void)
{
	gpio_reserve(GPIO_PORT107);
	gpio_reserve(GPIO_PORT108);

	if (!con7eb_detected) {
		gpio_leds_data.num_leds = 2;
		return;
	}

	if (IS_ENABLED(CONFIG_ARMADILLO840_LED_EXT1_PORT102)) {
		gpio_reserve(GPIO_PORT102);
		gpio_request_pulloff(portcr(102));
	}
	if (IS_ENABLED(CONFIG_ARMADILLO840_LED_EXT2_PORT59)) {
		gpio_reserve(GPIO_PORT59);
		gpio_request_pulloff(portcr(59));
	}
	if (IS_ENABLED(CONFIG_ARMADILLO840_LED_EXT3_PORT60)) {
		gpio_reserve(GPIO_PORT60);
		gpio_request_pulloff(portcr(60));
	}
	if (IS_ENABLED(CONFIG_ARMADILLO840_LED_EXT4_PORT10)) {
		gpio_reserve(GPIO_PORT10);
		gpio_request_pulloff(portcr(10));
	}
	if (IS_ENABLED(CONFIG_ARMADILLO840_LED_EXT5_PORT8)) {
		gpio_reserve(GPIO_PORT8);
		gpio_request_pulloff(portcr(8));
	}
	if (IS_ENABLED(CONFIG_ARMADILLO840_LED_EXT6_PORT7)) {
		gpio_reserve(GPIO_PORT7);
		gpio_request_pulloff(portcr(7));
	}
}

/* KEY */
#define GPIO_KEY(c, g, d, ...) \
	{ .code = c, .gpio = g, .desc = d, .active_low = 1, __VA_ARGS__ }

static struct gpio_keys_button gpio_buttons[] = {
#if IS_ENABLED(CONFIG_ARMADILLO840_GPIO_KEY_POWER_PORT97)
	GPIO_KEY(KEY_POWER,	GPIO_PORT97,	"SW1", .wakeup = 1),
#endif
#if IS_ENABLED(CONFIG_ARMADILLO840_GPIO_KEY_BACK_PORT98)
	GPIO_KEY(KEY_BACK,	GPIO_PORT98,	"SW2"),
#endif
#if IS_ENABLED(CONFIG_ARMADILLO840_GPIO_KEY_MENU_PORT99)
	GPIO_KEY(KEY_MENU,	GPIO_PORT99,	"SW3"),
#endif
#if IS_ENABLED(CONFIG_ARMADILLO840_GPIO_KEY_HOME_PORT100)
	GPIO_KEY(KEY_HOME,	GPIO_PORT100,	"SW4"),
#endif
};

static struct gpio_keys_platform_data gpio_key_info = {
	.buttons	= gpio_buttons,
	.nbuttons	= ARRAY_SIZE(gpio_buttons),
};

static struct platform_device gpio_keys_device = {
	.name   = "gpio-keys",
	.id     = -1,
	.dev    = {
		.platform_data  = &gpio_key_info,
	},
};

static void a840_gpio_keys_init(void)
{
	int i;

	if (!con7eb_detected)
		return;

	if (!gpio_key_info.nbuttons)
		return;

	for (i = 0; i < gpio_key_info.nbuttons; i++)
		gpio_reserve(gpio_buttons[i].gpio);

	platform_device_register(&gpio_keys_device);
}

/* GPIO */
static struct gpio con7_8_gpios[] = {
	{ GPIO_PORT195,			0,	"EXT_IO0" },
	{ GPIO_PORT196,			0,	"EXT_IO1" },
	{ GPIO_PORT23,			0,	"EXT_IO2" },
	{ GPIO_PORT21,			0,	"EXT_IO3" },
	{ GPIO_PORT160,			0,	"EXT_IO4" },
	{ GPIO_PORT197,			0,	"EXT_IO5" },
	{ GPIO_PORT198,			0,	"EXT_IO6" },
	{ GPIO_PORT194,			0,	"EXT_IO7" },
	{ GPIO_PORT193,			0,	"EXT_IO8" },
	{ GPIO_PORT62,			0,	"EXT_IO9" },
	{ GPIO_PORT63,			0,	"EXT_IO10" },
	{ GPIO_PORT64,			0,	"EXT_IO11" },
	{ GPIO_PORT65,			0,	"EXT_IO12" },
	{ GPIO_PORT61,			0,	"EXT_IO13" },
	{ GPIO_PORT165,			0,	"EXT_IO14" },
	{ GPIO_PORT164,			0,	"EXT_IO15" },
	{ GPIO_PORT202,			0,	"EXT_IO16" },
	{ GPIO_PORT102,			0,	"EXT_IO17" },
	{ GPIO_PORT59,			0,	"EXT_IO18" },
	{ GPIO_PORT60,			0,	"EXT_IO19" },
	{ GPIO_PORT172,			0,	"EXT_IO20" },
	{ GPIO_PORT173,			0,	"EXT_IO21" },
	{ GPIO_PORT4,			0,	"EXT_IO22" },
	{ GPIO_PORT3,			0,	"EXT_IO23" },
	{ GPIO_PORT2,			0,	"EXT_IO24" },
	{ GPIO_PORT0,			0,	"EXT_IO25" },
	{ GPIO_PORT1,			0,	"EXT_IO26" },
	{ GPIO_PORT66,			0,	"EXT_IO27" },
	{ GPIO_PORT67,			0,	"EXT_IO28" },
	{ GPIO_PORT68,			0,	"EXT_IO29" },
	{ GPIO_PORT69,			0,	"EXT_IO30" },
	{ GPIO_PORT70,			0,	"EXT_IO31" },
	{ GPIO_PORT71,			0,	"EXT_IO32" },
	{ GPIO_PORT72,			0,	"EXT_IO33" },
	{ GPIO_PORT73,			0,	"EXT_IO34" },
	{ GPIO_PORT74,			0,	"EXT_IO35" },
	{ GPIO_PORT75,			0,	"EXT_IO36" },
	{ GPIO_PORT97,			0,	"EXT_IO37" },
	{ GPIO_PORT98,			0,	"EXT_IO38" },
	{ GPIO_PORT99,			0,	"EXT_IO39" },
	{ GPIO_PORT100,			0,	"EXT_IO40" },
	{ GPIO_PORT13,			0,	"EXT_IO42" },
	{ GPIO_PORT12,			0,	"EXT_IO43" },
	{ GPIO_PORT9,			0,	"EXT_IO44" },
	{ GPIO_PORT5,			0,	"EXT_IO45" },
	{ GPIO_PORT20,			0,	"EXT_IO46" },
	{ GPIO_PORT10,			0,	"EXT_IO47" },
	{ GPIO_PORT8,			0,	"EXT_IO48" },
	{ GPIO_PORT7,			0,	"EXT_IO49" },
	{ GPIO_PORT40,			0,	"EXT_IO50" },
	{ GPIO_PORT41,			0,	"EXT_IO51" },
	{ GPIO_PORT42,			0,	"EXT_IO52" },
	{ GPIO_PORT43,			0,	"EXT_IO53" },
	{ GPIO_PORT44,			0,	"EXT_IO54" },
	{ GPIO_PORT45,			0,	"EXT_IO55" },
	{ GPIO_PORT46,			0,	"EXT_IO56" },
	{ GPIO_PORT47,			0,	"EXT_IO57" },
	{ GPIO_PORT48,			0,	"EXT_IO58" },
	{ GPIO_PORT49,			0,	"EXT_IO59" },
	{ GPIO_PORT50,			0,	"EXT_IO60" },
	{ GPIO_PORT51,			0,	"EXT_IO61" },
	{ GPIO_PORT52,			0,	"EXT_IO62" },
	{ GPIO_PORT53,			0,	"EXT_IO63" },
	{ GPIO_PORT54,			0,	"EXT_IO64" },
	{ GPIO_PORT55,			0,	"EXT_IO65" },
	{ GPIO_PORT56,			0,	"EXT_IO66" },
	{ GPIO_PORT57,			0,	"EXT_IO67" },
	{ GPIO_PORT58,			0,	"EXT_IO68" },
	{ GPIO_PORT24,			0,	"EXT_IO69" },
	{ GPIO_PORT25,			0,	"EXT_IO70" },
	{ GPIO_PORT26,			0,	"EXT_IO71" },
	{ GPIO_PORT178,			0,	"EXT_IO72" },
	{ GPIO_PORT179,			0,	"EXT_IO73" },
	{ GPIO_PORT180,			0,	"EXT_IO74" },
	{ GPIO_PORT181,			0,	"EXT_IO75" },
	{ GPIO_PORT182,			0,	"EXT_IO76" },
	{ GPIO_PORT34,			0,	"EXT_IO77" },
	{ GPIO_PORT33,			0,	"EXT_IO78" },
	{ GPIO_PORT32,			0,	"EXT_IO79" },
	{ GPIO_PORT31,			0,	"EXT_IO80" },
	{ GPIO_PORT30,			0,	"EXT_IO81" },
	{ GPIO_PORT29,			0,	"EXT_IO82" },
	{ GPIO_PORT28,			0,	"EXT_IO83" },
	{ GPIO_PORT27,			0,	"EXT_IO84" },
	{ GPIO_PORT35,			0,	"EXT_IO85" },
	{ GPIO_PORT38,			0,	"EXT_IO86" },
	{ GPIO_PORT37,			0,	"EXT_IO87" },
	{ GPIO_PORT39,			0,	"EXT_IO88" },
	{ GPIO_PORT36,			0,	"EXT_IO89" },
	{ GPIO_PORT158,			0,	"EXT_IO90" },
	{ GPIO_PORT159,			0,	"EXT_IO91" },
	{ GPIO_PORT199,			0,	"EXT_IO92" },
	{ GPIO_PORT94,			0,	"EXT_IO93" },
	{ GPIO_PORT93,			0,	"EXT_IO94" },
	{ GPIO_PORT22,			0,	"EXT_IO95" },
};

static void activate_ext_gpios(const struct gpio *array, size_t num)
{
	int i;

	for (i = 0; i < num; i++, array++) {
		if (is_gpio_reserved(array->gpio))
			continue;

		if (!gpio_request(array->gpio, array->label))
			gpio_export(array->gpio, 1);
	}
}

static void a840_ext_gpio_init(void)
{
	activate_ext_gpios(con7_8_gpios, ARRAY_SIZE(con7_8_gpios));
}

/* I2C */
static struct i2c_board_info i2c0_devices[] = {
	{
		I2C_BOARD_INFO("st1232-ts", 0x55),
		.irq	= irq_pin(4), /* IRQ4A */
	},
	{
		I2C_BOARD_INFO("wm8978", 0x1a),
	},
};

static struct i2c_board_info i2c2_devices[] = {
	{
		I2C_BOARD_INFO("s35390a", 0x30),
		.irq	= irq_pin(11),
		.flags	= RTC_S35390A_FLAG_INT1_EN,
	},
};

static struct i2c_gpio_platform_data __maybe_unused i2c2_gpio_data = {
	.scl_pin	= GPIO_PORT114,
	.sda_pin	= GPIO_PORT106,
	.udelay		= 5,
};

static struct platform_device __maybe_unused i2c2_gpio_device = {
	.name		= "i2c-gpio",
	.id		= 2,
	.dev		= {
		.platform_data	= &i2c2_gpio_data,
	},
};

static struct i2c_gpio_platform_data __maybe_unused i2c3_gpio_data = {
	.scl_pin	= GPIO_PORT194,
	.sda_pin	= GPIO_PORT193,
	.udelay		= 5,
};

static struct platform_device __maybe_unused i2c3_gpio_device = {
	.name		= "i2c-gpio",
	.id		= 3,
	.dev		= {
		.platform_data	= &i2c3_gpio_data,
	},
};

static void a840_i2c_init(void)
{
	if (con7eb_detected)
		i2c_register_board_info(0, i2c0_devices,
					ARRAY_SIZE(i2c0_devices));

	gpio_reserve(GPIO_PORT114);
	gpio_reserve(GPIO_PORT106);

	/* PORT104 is connected to INT1 of S-35390A */
	gpio_request(GPIO_FN_IRQ11, NULL);
	gpio_request(GPIO_PORT104, NULL);
	gpio_direction_input(GPIO_PORT104);
	gpio_request_pulloff(portcr(104));

	i2c_register_board_info(2, i2c2_devices, ARRAY_SIZE(i2c2_devices));
	platform_device_register(&i2c2_gpio_device);

	if (IS_ENABLED(CONFIG_ARMADILLO840_I2C_GPIO3_194_193)) {
		gpio_reserve(GPIO_PORT194);
		gpio_reserve(GPIO_PORT193);

		platform_device_register(&i2c3_gpio_device);
	}
}

/* Watchdog Timer */
static struct resource wdt_resources[] = {
	{
		.start	= 0xe6020000,
		.end	= 0xe602000c - 1,
		.flags	= IORESOURCE_MEM,
	},
	{
		.start	= gic_spi(0),
		.flags	= IORESOURCE_IRQ,
	},
};

static struct platform_device wdt_device = {
	.name		= "sh_mobile_wdt",
	.id		= 0,
	.resource	= wdt_resources,
	.num_resources	= ARRAY_SIZE(wdt_resources),
};

/*
 * board devices
 */
static struct platform_device *a840_devices[] __initdata = {
	&wdt_device,
	&flash_device,
	&usb_ehci_device,
	&usb_ohci_device,
	&sh_eth_device,
	&cec_device,
	&gpio_leds_device,
	&sgx540_device,
	&acm_device,
};

static struct sh_timer_config cmt10_platform_data = {
	.name			= "CMT10",
	.channel_offset		= 0x10,
	.timer_bit		= 0,
	.clocksource_rating	= 125,
};

static struct resource cmt10_resources[] = {
	[0] = {
		.name	= "CMT10",
		.start	= 0xe6138010,
		.end	= 0xe613801b,
		.flags	= IORESOURCE_MEM,
	},
	[1] = {
		.start	= gic_spi(58),
		.flags	= IORESOURCE_IRQ,
	},
};

static struct platform_device cmt10_device = {
	.name		= "sh_cmt_simple",
	.id		= 10,
	.dev = {
		.platform_data	= &cmt10_platform_data,
	},
	.resource	= cmt10_resources,
	.num_resources	= ARRAY_SIZE(cmt10_resources),
};

static struct sh_timer_config cmt14_platform_data = {
	.name			= "CMT14",
	.channel_offset		= 0x50,
	.timer_bit		= 4,
	.clockevent_rating	= 125,
};

static struct resource cmt14_resources[] = {
	[0] = {
		.name	= "CMT14",
		.start	= 0xe6138050,
		.end	= 0xe613805b,
		.flags	= IORESOURCE_MEM,
	},
	[1] = {
		.start	= gic_spi(62),
		.flags	= IORESOURCE_IRQ,
	},
};

static struct platform_device cmt14_device = {
	.name		= "sh_cmt_simple",
	.id		= 14,
	.dev = {
		.platform_data	= &cmt14_platform_data,
	},
	.resource	= cmt14_resources,
	.num_resources	= ARRAY_SIZE(cmt14_resources),
};

static struct platform_device *a840_early_devices[] __initdata = {
	&cmt10_device,
	&cmt14_device,
};

static void __init a840_clock_init(void)
{
	struct clk *system	= clk_get(NULL, "system_clk");
	struct clk *usb24s	= clk_get(NULL, "usb24s");
	struct clk *fsibck	= clk_get(NULL, "fsibck");
	struct clk *fsib	= clk_get(&fsi_device.dev, "ickb");

	if (IS_ERR(system)	||
	    IS_ERR(usb24s)	||
	    IS_ERR(fsibck)	||
	    IS_ERR(fsib)) {
		pr_err("armadillo840 board clock init failed\n");
		goto clock_error;
	}

	/* usb24s use extal1 (= system) clock (= 24MHz) */
	clk_set_parent(usb24s, system);

	/* FSIBCK is 12.288MHz, and it is parent of FSI-B */
	clk_set_parent(fsib, fsibck);
	clk_set_rate(fsibck, 12288000);
	clk_set_rate(fsib,   12288000);
	clk_enable(fsib);

clock_error:
	if (!IS_ERR(system))
		clk_put(system);
	if (!IS_ERR(usb24s))
		clk_put(usb24s);
	if (!IS_ERR(fsibck))
		clk_put(fsibck);
	if (!IS_ERR(fsib))
		clk_put(fsib);
}

static void __init a840_early_clock_init(void)
{
	struct clk *xtal1 = clk_get(NULL, "extal1");

	/* armadillo840 extal1 is 24MHz */
	clk_set_rate(xtal1, 24000000);
}

#define RESCNT2		0xe6188020
static void a840_restart(char str, const char *cmd)
{
	/* Soft power-on reset
	 *   All power areas including C5, and D4 are initilized. */
	__raw_writel(__raw_readl(RESCNT2) | (1 << 31), RESCNT2);
}

/*
 * board init
 */
static void __init a840_init(void)
{
	r8a7740_pinmux_init();
	r8a7740_meram_workaround();
	r8a7740_sgx540_workaround();

	/* if direction of PORT74 is INPUT, PORT74 is detection signal
	 * of extention board. if we detect extention board,
	 * we register for some platform device and do pinmux.
	 * if PORT74 is Low, extension board is connected to armadillo-840. */
	if (IS_ENABLED(CONFIG_ARMADILLO840_CON7EB_LCD)) {
		gpio_request(GPIO_PORT74, NULL);
		gpio_direction_input(GPIO_PORT74);
		con7eb_detected = !gpio_get_value(GPIO_PORT74);
		gpio_free(GPIO_PORT74);
		pr_info("CON7: %s extension board found.\n",
			con7eb_detected ? "LCD" : "no");
	} else {
		/* this is implementation dependent.
		 *
		 * some extension board does not have the detection signal.
		 * if your extension board has detection signal,
		 * use gpio_get_value() to detect it.
		 * otherwise assign 1 to force detected, or 0. */
		con7eb_detected = 1;
	}

	/* init peripheral modules */
	a840_scifx_init();
	a840_eth_init();
	a840_usb_init();
	a840_sdhi_init();
	a840_lcdc_init();
	a840_camera_init();
	a840_audio_init();
	a840_i2c_init();
	a840_msiof_init();
	a840_gpio_keys_init();
	a840_led_init();
	a840_ext_gpio_init();

	if (IS_ENABLED(CONFIG_CACHE_L2X0)) {
		/* Early BRESP enable, Shared attribute override enable,
		 * 32K*8way */
		l2x0_init((void __iomem *)0xf0002000, 0x40440000, 0x82000fff);
	}

	r8a7740_add_standard_devices();

	platform_add_devices(a840_devices,
			     ARRAY_SIZE(a840_devices));

	a840_clock_init();
}

static void __init a840_earlytimer_init(void)
{
	r8a7740_clock_init(MD_CK0);
	a840_early_clock_init();
	shmobile_earlytimer_init();
}

static void __init a840_add_early_devices(void)
{
	r8a7740_add_early_devices();
	early_platform_add_devices(a840_early_devices,
				   ARRAY_SIZE(a840_early_devices));

	/* override timer setup with board-specific code */
	shmobile_timer.init = a840_earlytimer_init;
}

MACHINE_START(ARMADILLO840, "armadillo840")
	.map_io		= r8a7740_map_io,
	.init_early	= a840_add_early_devices,
	.init_irq	= r8a7740_init_irq,
	.handle_irq	= gic_handle_irq,
	.init_machine	= a840_init,
	.timer		= &shmobile_timer,
	.restart	= a840_restart,
MACHINE_END
